# ruff: noqa: F403, F405, E402, E501

from pydantic import BaseModel
from typing import Optional
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *


class Reward:
    """ """

    def __init__(self, r):
        self.r = r

    def account_reward(self):
        r = self.r
        if r.payday_account_reward:
            return r.payday_account_reward.account
        else:
            return None

    def pool_reward(self):
        r = self.r

        if r.payday_pool_reward:
            return r.payday_pool_reward.pool_owner
        else:
            return None


class NotificationServices(str, Enum):
    telegram = "telegram"
    email = "email"


class NotificationService(BaseModel):
    enabled: Optional[bool] = None
    limit: Optional[microCCD] = None
    exclude_exchange_to_exchange: Optional[bool] = None


class NotificationPreferences(BaseModel):
    telegram: Optional[NotificationService] = None
    email: Optional[NotificationService] = None


class ContractNotificationPreferences(BaseModel):
    # this is a list of methods that can be individually enabled
    contract_update_issued: Optional[dict[str, NotificationPreferences]] = None


class AccountNotificationPreferences(BaseModel):
    validator_commission_changed: Optional[NotificationPreferences] = None
    validator_primed_for_suspension: Optional[NotificationPreferences] = None
    validator_suspended: Optional[NotificationPreferences] = None
    delegation_configured: Optional[NotificationPreferences] = None
    payday_account_reward: Optional[NotificationPreferences] = None
    account_transfer: Optional[NotificationPreferences] = None
    transferred_with_schedule: Optional[NotificationPreferences] = None
    data_registered: Optional[NotificationPreferences] = None
    module_deployed: Optional[NotificationPreferences] = None
    contract_initialized: Optional[NotificationPreferences] = None
    encrypted_amount_transferred: Optional[NotificationPreferences] = None
    transferred_to_encrypted: Optional[NotificationPreferences] = None
    transferred_to_public: Optional[NotificationPreferences] = None

    credential_keys_updated: Optional[NotificationPreferences] = None
    credentials_updated: Optional[NotificationPreferences] = None

    # if the account is delegating, get notified if your validator configures

    token_event: Optional[NotificationPreferences] = None


class ValidatorNotificationPreferences(BaseModel):
    block_validated: Optional[NotificationPreferences] = None
    validator_configured: Optional[NotificationPreferences] = None
    payday_pool_reward: Optional[NotificationPreferences] = None
    delegation_configured: NotificationPreferences = None
    validator_running_behind: Optional[NotificationPreferences] = None
    # baker_configured: Optional[NotificationPreferences] = None
    # block_baked_by_baker: Optional[NotificationPreferences] = None
    validator_primed_for_suspension: Optional[NotificationPreferences] = None
    validator_suspended: Optional[NotificationPreferences] = None
    validator_missed_rounds: Optional[NotificationPreferences] = None


class OtherNotificationPreferences(BaseModel):
    validator_commission_changed: Optional[NotificationPreferences] = None
    protocol_update: Optional[NotificationPreferences] = None
    add_anonymity_revoker_update: Optional[NotificationPreferences] = None
    add_identity_provider_update: Optional[NotificationPreferences] = None
    account_transfer: Optional[NotificationPreferences] = None
    transferred_with_schedule: Optional[NotificationPreferences] = None
    domain_name_minted: Optional[NotificationPreferences] = None
    module_deployed: Optional[NotificationPreferences] = None
    contract_initialized: Optional[NotificationPreferences] = None
    account_created: Optional[NotificationPreferences] = None
    # lowered_stake: Optional[NotificationPreferences] = None
    validator_lowered_stake: Optional[NotificationPreferences] = None
    validator_suspended: Optional[NotificationPreferences] = None
    validator_primed_for_suspension: Optional[NotificationPreferences] = None
    validator_node_version_upgrade: Optional[NotificationPreferences] = None
    token_creation: Optional[NotificationPreferences] = None


class AccountForUser(BaseModel):
    account_index: Optional[int] = None
    account_id: Optional[CCD_AccountAddress] = None
    # only set if targeting a validator (and not passive)
    delegation_target: Optional[CCD_AccountIndex] = None
    label: Optional[str] = None
    account_notification_preferences: Optional[AccountNotificationPreferences] = None
    validator_notification_preferences: Optional[ValidatorNotificationPreferences] = (
        None
    )


class ContractForUser(BaseModel):
    contract: Optional[CCD_ContractAddress] = None
    contract_name: Optional[str] = None
    label: Optional[str] = None
    contract_notification_preferences: Optional[ContractNotificationPreferences] = None


class UserV2(BaseModel):
    telegram_chat_id: Optional[int] = None
    email_address: Optional[str] = None
    token: str
    username: Optional[str] = None
    first_name: Optional[str] = None
    language_code: Optional[str] = None
    accounts: Optional[dict] = {}
    contracts: Optional[dict] = {}
    other_notification_preferences: Optional[OtherNotificationPreferences] = None
    last_modified: Optional[dt.datetime] = None
    last_seen_on_site: Optional[dt.datetime] = None



# ruff: noqa: F403, F405, E402
from __future__ import annotations

import datetime as dt
from datetime import timezone
from enum import Enum
from typing import Any, Protocol, Union
from decimal import Decimal
import base58
import cbor2
from google._upb._message import RepeatedCompositeContainer
from google.protobuf.json_format import MessageToDict

from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *
from ccdexplorer_fundamentals.GRPCClient.kernel_pb2 import *
from ccdexplorer_fundamentals.GRPCClient.protocol_level_tokens_pb2 import *
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *


class OpenStatusEnum(Enum):
    openForAll = 0
    closedForNew = 1
    closedForAll = 2


class TransactionType(Enum):
    DEPLOY_MODULE = 0
    INIT_CONTRACT = 1
    UPDATE = 2
    TRANSFER = 3
    ADD_BAKER = 4
    REMOVE_BAKER = 5
    UPDATE_BAKER_STAKE = 6
    UPDATE_BAKER_RESTAKE_EARNINGS = 7
    UPDATE_BAKER_KEYS = 8
    UPDATE_CREDENTIAL_KEYS = 9
    ENCRYPTED_AMOUNT_TRANSFER = 10
    TRANSFER_TO_ENCRYPTED = 11
    TRANSFER_TO_PUBLIC = 12
    TRANSFER_WITH_SCHEDULE = 13
    UPDATE_CREDENTIALS = 14
    REGISTER_DATA = 15
    TRANSFER_WITH_MEMO = 16
    ENCRYPTED_AMOUNT_TRANSFER_WITH_MEMO = 17
    TRANSFER_WITH_SCHEDULE_AND_MEMO = 18
    CONFIGURE_BAKER = 19
    CONFIGURE_DELEGATION = 20


class Mixin(Protocol):
    # These types should be encoded to HEX
    bytes_to_hex_types = [
        bytes,
        BlockHash,
        TransactionHash,
        CredentialRegistrationId,
        StateHash,
        Memo,
        ModuleRef,
        Commitment,
        EncryptionKey,
        EncryptedAmount,
        BakerElectionVerifyKey,
        BakerSignatureVerifyKey,
        BakerAggregationVerifyKey,
        RegisteredData,
        IpInfo.IpVerifyKey,
        IpInfo.IpCdiVerifyKey,
        ArInfo.ArPublicKey,
        LeadershipElectionNonce,
        Parameter,
        Sha256Hash,
        ContractStateV0,
        VersionedModuleSource.ModuleSourceV0,
        VersionedModuleSource.ModuleSourceV1,
        QuorumSignature,
        TimeoutSignature,
        BlockSignature,
        SuccessorProof,
        FinalizationCommitteeHash,
        Cbor,
        TokenModuleRef,
    ]
    # These types should be shown as is (from the .value property)
    value_property_types = [
        AbsoluteBlockHeight,
        Amount,
        AccountThreshold,
        AccountIndex,
        ArInfo.ArIdentity,
        ArThreshold,
        CredentialsPerBlockLimit,
        BlockHeight,
        BlockItemSummary.TransactionIndex,
        DurationSeconds,
        Duration,
        Energy,
        RewardPeriodLength,
        Epoch,
        InitName,
        IpIdentity,
        IdentityProviderIdentity,
        GenesisIndex,
        ReceiveName,
        SequenceNumber,
        SignatureThreshold,
        Slot,
        TransactionTime,
        UpdateKeysThreshold,
        Round,
        FinalizerIndex,
        TokenId,
    ]
    # These types need special attention for conversion
    remaining_types = [
        AccountAddress,
        Address,
        AccountVerifyKey,
        BakerId,
        ChainArData,
        ContractAddress,
        DelegatorId,
        Empty,
        Timestamp,
        ElectionDifficulty,
        str,
        int,
        bool,
        float,
        TokenHolder,
    ]

    # all types that are covered by the `convertType` method.
    simple_types = bytes_to_hex_types + value_property_types + remaining_types

    def get_key_value_from_descriptor(self, descriptor, the_list):
        return descriptor.name, getattr(the_list, descriptor.name)

    def generate_account_identifier_input_from(self, hex_address: str):
        try:
            bin_value = base58.b58decode_check(hex_address)[1:]
            address = AccountAddress(value=bin_value)
            account = AccountIdentifierInput(address=address)
            return account
        except:  # noqa : F403, E722
            return None

    def generate_account_identifier_input_from_account_index(
        self, account_index: CCD_AccountIndex
    ):
        try:
            account = AccountIdentifierInput(
                account_index=AccountIndex(value=account_index)
            )
            return account
        except:  # noqa: E722
            return None

    def generate_epoch_request_from_genesis(
        self, genesis_index: int, epoch: int
    ) -> EpochRequest:
        relative_epoch = EpochRequest.RelativeEpoch(
            genesis_index=GenesisIndex(value=genesis_index),
            epoch=Epoch(value=epoch),
        )
        return EpochRequest(relative_epoch=relative_epoch)

    def generate_consensus_detailed_status_query(self):
        return ConsensusDetailedStatusQuery()

    def valueIsEmpty(self, value, key=None, message=None):
        if isinstance(value, int):
            return value is None
        else:
            if type(value) is RepeatedCompositeContainer:
                return False
            else:
                if hasattr(value, "DESCRIPTOR"):
                    if message:
                        lll = list(message)
                        key_in_message = key in lll[0].__str__()
                        if key_in_message:
                            # special case for baker_id = 0
                            # (doesn't get transmitted in a message)
                            if key in ["baker_removed"]:
                                return False
                            else:
                                if MessageToDict(value) != {}:
                                    return False
                                else:
                                    return True
                        return MessageToDict(value) == {}
                    else:
                        return MessageToDict(value) == {}
                else:  # pragma: no cover
                    return False

    def generate_block_hash_input_from(
        self, block_input: Union[str, int]
    ) -> BlockHashInput:
        if isinstance(block_input, str):
            if block_input == "last_final":
                return BlockHashInput(last_final={})
            else:
                return BlockHashInput(given=BlockHash(value=bytes.fromhex(block_input)))
        elif isinstance(block_input, int):
            absoluteBlockHeight = AbsoluteBlockHeight(value=block_input)
            return BlockHashInput(absolute_height=absoluteBlockHeight)

    def generate_invoke_instance_request_from(
        self,
        contract_index: int,
        contract_sub_index: int,
        block_hash_input: str,
        receive_name: str,
        parameter_bytes: bytes,
    ):
        # parameter_bytes = self.get_bytes_parameter_from_hex(parameter_hex)
        # if we include a zero subindex, grpc will complain about
        # LengthDelimited vs VarInt.
        if contract_sub_index == 0:
            instance = ContractAddress(index=contract_index)
        else:
            instance = ContractAddress(
                index=contract_index,
                subindex=contract_sub_index,
            )

        return InvokeInstanceRequest(
            block_hash=block_hash_input,
            amount=Amount(value=0),
            entrypoint=ReceiveName(value=receive_name),
            parameter=Parameter(value=parameter_bytes),
            energy=Energy(value=100_000_000_000),
            instance=instance,
        )

    def generate_instance_info_request_from(
        self, contract_index: int, contract_sub_index: int, hex_block_hash: str
    ):
        return InstanceInfoRequest(
            block_hash=self.generate_block_hash_input_from(hex_block_hash),
            address=ContractAddress(
                **{"index": contract_index, "subindex": contract_sub_index}
            ),
        )

    def generate_module_source_request_from(
        self, module_ref: CCD_ModuleRef, hex_block_hash: str
    ):
        return ModuleSourceRequest(
            block_hash=self.generate_block_hash_input_from(hex_block_hash),
            module_ref=ModuleRef(value=bytes.fromhex(module_ref)),
        )

    def convertList(self, message):
        entries = []

        for list_entry in message:
            entries.append(self.convertType(list_entry))

        return entries

    def convertType(self, value) -> Any:
        # these types have a property `value` that we need to return unmodified
        if type(value) in self.value_property_types:
            return value.value

        if isinstance(value, bytes):
            return value.hex()
        # these types have a property `value` that we need to encode as HEX>
        elif type(value) in self.bytes_to_hex_types:
            return value.value.hex()

        elif type(value) in [int, bool, str, float]:
            return value

        elif type(value) is ElectionDifficulty:
            return value.value.parts_per_hundred_thousand / 100_000

        elif type(value) is AmountFraction:
            return value.parts_per_hundred_thousand / 100_000

        elif type(value) is Empty:
            return None  # pragma: no cover
        elif type(value) is BakerEvent.BakerResumed:
            return value.baker_id.value

        elif type(value) is Address:
            if MessageToDict(value.account) == {}:
                return CCD_Address(
                    **{"contract": self.convertContractAddress(value.contract)}
                )
            else:
                return CCD_Address(
                    **{"account": self.convertAccountAddress(value.account)}
                )

        elif type(value) is AccountAddress:
            return self.convertAccountAddress(value)

        elif type(value) is BakerId:
            return value.value
            # else:
            #     return None

        elif type(value) is ContractAddress:
            return self.convertContractAddress(value)

        elif type(value) is AccountVerifyKey:
            return value.ed25519_key.hex()

        elif type(value) is ChainArData:
            return value.enc_id_cred_pub_share.hex()

        elif type(value) is Timestamp:
            if MessageToDict(value) == {}:
                pass
            else:
                return dt.datetime.fromtimestamp(
                    int(MessageToDict(value)["value"]) / 1_000, tz=timezone.utc
                )

        elif type(value) is DelegatorId:
            return value.id.value

        elif type(value) is TokenHolder:
            return CCD_TokenHolder(account=self.convertAccountAddress(value.account))
        # elif type(value) is TokenModuleRef:
        #     return CCD_TokenModuleRef(value=value.value)
        elif type(value) is TokenId:
            return value.value  # CCD_TokenId(value=value.value)

    def convertContractAddress(self, value: ContractAddress) -> CCD_ContractAddress:
        return CCD_ContractAddress(**{"index": value.index, "subindex": value.subindex})

    def convertAccountAddress(self, value: AccountAddress) -> CCD_AccountAddress:
        return base58.b58encode_check(b"\x01" + value.value).decode()

    def convertAmount(self, value: Amount) -> microCCD:
        return value.value

    def convertDelegationTarget(self, message) -> CCD_DelegationTarget:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) is Empty:
                pass
            if type(value) is BakerId:
                if self.valueIsEmpty(value):
                    result["passive_delegation"] = True
                else:
                    # result['passive_delegation'] = False
                    result["baker"] = self.convertType(value)

        return CCD_DelegationTarget(**result)

    def convertCommissionRates(self, value) -> CCD_CommissionRates:
        result = {}
        for descriptor in value.DESCRIPTOR.fields:
            key, val = self.get_key_value_from_descriptor(descriptor, value)
            result[key] = val.parts_per_hundred_thousand / 100_000
        return CCD_CommissionRates(**result)

    def convertValidatorScoreParameters(self, message) -> CCD_ValidatorScoreParameters:
        return CCD_ValidatorScoreParameters(
            **{"maximum_missed_rounds": message.maximum_missed_rounds}
        )

    def convertRelease(self, message) -> CCD_ReleaseSchedule:
        resulting_dict = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if key == "schedules":
                schedule = []
                for entry in value:
                    entry_dict = {}
                    for descriptor in entry.DESCRIPTOR.fields:
                        key, value = self.get_key_value_from_descriptor(
                            descriptor, entry
                        )

                        if key == "transactions":
                            entry_dict[key] = self.convertList(value)

                        elif type(value) is Timestamp:
                            entry_dict[key] = self.convertType(value)

                        elif type(value) is Amount:
                            entry_dict[key] = self.convertType(value)

                    schedule.append(entry_dict)
                resulting_dict["schedules"] = schedule

            elif type(value) is Amount:
                resulting_dict[key] = self.convertType(value)

        return CCD_ReleaseSchedule(**resulting_dict)

    def convertCoolDowns(self, message) -> list[CCD_Cooldown]:
        resulting_dict = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            # if key == "schedules":
            #     schedule = []
            #     for entry in value:
            #         entry_dict = {}
            #         for descriptor in entry.DESCRIPTOR.fields:
            #             key, value = self.get_key_value_from_descriptor(
            #                 descriptor, entry
            #             )

            #             if key == "transactions":
            #                 entry_dict[key] = self.convertList(value)

            #             elif type(value) is Timestamp:
            #                 entry_dict[key] = self.convertType(value)

            #             elif type(value) is Amount:
            #                 entry_dict[key] = self.convertType(value)

            #         schedule.append(entry_dict)
            #     resulting_dict["schedules"] = schedule

            # elif type(value) is Amount:
            #     resulting_dict[key] = self.convertType(value)

        return CCD_Cooldown(**resulting_dict)

    def convertAccountIndex(self, message) -> CCD_AccountIndex:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            result[key] = self.convertType(value)

        return CCD_AccountIndex(**result)

    def convertAccountPending(self, message) -> CCD_AccountPending:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in [AccountIndex, Timestamp]:
                result[key] = self.convertType(value)

        return CCD_AccountPending(**result)

    def convertWinningBaker(self, message) -> CCD_WinningBaker:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            result[key] = self.convertType(value)

        return CCD_WinningBaker(**result)

    def convertArInfo(self, message) -> CCD_ArInfo:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in [ArInfo.ArIdentity, ArInfo.ArPublicKey]:
                result[key] = self.convertType(value)

            elif type(value) is Description:
                result[key] = self.convertTypeWithSingleValues(value)

        return CCD_ArInfo(**result)

    def convertIpInfo(self, message) -> CCD_IpInfo:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in [IpIdentity, IpInfo.IpVerifyKey, IpInfo.IpCdiVerifyKey]:
                result[key] = self.convertType(value)

            elif type(value) is Description:
                result[key] = self.convertTypeWithSingleValues(value)

        return CCD_IpInfo(**result)

    def convertBakerPoolInfo(self, message) -> CCD_BakerPoolInfo:
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            if key == "open_status":
                result[key] = CCD_OpenStatusTranslation(value).name

            elif type(value) is CommissionRates:
                result[key] = self.convertCommissionRates(value)

        return CCD_BakerPoolInfo(**result)

    def convertPoolCurrentPaydayInfo(self, message) -> CCD_CurrentPaydayStatus:
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            result[key] = self.convertType(value)

        return CCD_CurrentPaydayStatus(**result)

    def convertInclusiveRange(self, message) -> CCD_InclusiveRangeAmountFraction:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) is AmountFraction:
                result[key] = self.convertType(value)

        return CCD_InclusiveRangeAmountFraction(**result)

    def convertTypeWithSingleValues(self, message):
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) is InclusiveRangeAmountFraction:
                result[key] = self.convertTypeWithSingleValues(value)

            elif type(value) is Ratio:
                result[key] = CCD_Ratio(
                    **{
                        "numerator": str(value.numerator),
                        "denominator": str(value.denominator),
                    }
                )

            elif type(value) is RewardPeriodLength:
                result[key] = self.convertType(value.value)

            elif type(value) is MintRate:
                result[key] = CCD_MintRate(
                    **{"mantissa": value.mantissa, "exponent": value.exponent}
                )

            elif type(value) is TokenId:
                result[key] = value.value
            elif type(value) is BakerId:
                if descriptor.json_name in MessageToDict(message):
                    result[key] = self.convertType(value)
            else:
                result[key] = self.convertType(value)

        return result

    def convertPendingChange_Reduce_Remove(
        self, message
    ) -> CCD_StakePendingChange_Remove:
        return self.convertType(message)

    def convertPendingChange(self, message) -> CCD_StakePendingChange:
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if message.WhichOneof("change") == "reduce" and key == "reduce":
                result[key] = self.convertPendingChange_Reduce_Remove(value)
            elif message.WhichOneof("change") == "remove" and key == "remove":
                result[key] = self.convertPendingChange_Reduce_Remove(value)

        return CCD_StakePendingChange(**result)

    def convertPoolPendingChange_Reduce_Remove(
        self, message
    ) -> CCD_BakerStakePendingChange_Remove:
        return self.convertTypeWithSingleValues(message)

    def convertPoolPendingChange(self, message) -> CCD_BakerStakePendingChange:
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if message.WhichOneof("change") == "reduce" and key == "reduce":
                result[key] = self.convertPoolPendingChange_Reduce_Remove(value)
            elif message.WhichOneof("change") == "remove" and key == "remove":
                result[key] = self.convertPoolPendingChange_Reduce_Remove(value)

        return CCD_BakerStakePendingChange(**result)

    def convertExchangeRate(self, message) -> CCD_ExchangeRate:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if MessageToDict(value) == {}:
                    pass
                else:
                    if type(value) is Ratio:
                        result[key] = CCD_Ratio(
                            **{
                                "numerator": value.numerator,
                                "denominator": value.denominator,
                            }
                        )

        return CCD_ExchangeRate(**result)

    def convertUpdatePublicKeys(self, message) -> list[CCD_UpdatePublicKey]:
        keys = []

        for entry in message:
            for descriptor in entry.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, entry)

                # TODO: this needs to be decoded to hex still
                if type(value) in self.simple_types:
                    keys.append(self.convertType(value))

        return keys

    def convertHigherLevelKeys(self, message) -> CCD_HigherLevelKeys:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if key == "keys":
                    result[key] = self.convertUpdatePublicKeys(value)

                elif type(value) in self.simple_types:
                    result[key] = self.convertType(value)

        return CCD_HigherLevelKeys(**result)

    def convertAccessPublicKeys(self, message) -> list[CCD_UpdatePublicKey]:
        keys = []

        for entry in message:
            for descriptor in entry.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, entry)

                if type(value) in self.simple_types:
                    keys.append(self.convertType(value))

        return keys

    def convertAccessStructure(self, message) -> CCD_AccessStructure:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if self.valueIsEmpty(value):
                    pass
                else:
                    if key == "access_public_keys":
                        result[key] = self.convertAccessPublicKeys(value)

                    elif type(value) in self.simple_types:
                        result[key] = self.convertType(value)

        return CCD_AccessStructure(**result)

    def convertAuthorizationsV0(self, message) -> CCD_AuthorizationsV0:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if self.valueIsEmpty(value):
                    pass
                else:
                    if key == "keys":
                        result[key] = self.convertUpdatePublicKeys(value)

                    elif type(value) is AccessStructure:
                        result[key] = self.convertAccessStructure(value)

                    elif type(value) in self.simple_types:
                        result[key] = self.convertType(value)

        return CCD_AuthorizationsV0(**result)

    def convertAuthorizationsV1(self, message) -> CCD_AuthorizationsV1:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if self.valueIsEmpty(value):
                    pass
                else:
                    if type(value) is AuthorizationsV0:
                        result[key] = self.convertAuthorizationsV0(value)

                    elif type(value) is AccessStructure:
                        result[key] = self.convertAccessStructure(value)

                    elif type(value) in self.simple_types:
                        result[key] = self.convertType(value)

        return CCD_AuthorizationsV1(**result)

    def convertLevel1Update(self, message) -> CCD_Level1Update:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if MessageToDict(value) == {}:
                    pass
                else:
                    if type(value) is HigherLevelKeys:
                        result[key] = self.convertHigherLevelKeys(value)

                    elif type(value) is AuthorizationsV0:
                        result[key] = self.convertAuthorizationsV0(value)

                    elif type(value) is AuthorizationsV1:
                        result[key] = self.convertAuthorizationsV1(value)

        return CCD_Level1Update(**result)

    def convertRootUpdate(self, message) -> CCD_RootUpdate:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if MessageToDict(value) == {}:
                    pass
                else:
                    if type(value) is HigherLevelKeys:
                        result[key] = self.convertHigherLevelKeys(value)

                    elif type(value) is AuthorizationsV0:
                        result[key] = self.convertAuthorizationsV0(value)

                    elif type(value) is AuthorizationsV1:
                        result[key] = self.convertAuthorizationsV1(value)

        return CCD_RootUpdate(**result)

    def converCommissionRanges(self, message) -> CCD_CommissionRanges:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_CommissionRanges(**result)

    def convertPoolParametersCpv1(self, message) -> CCD_PoolParametersCpv1:
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if MessageToDict(value) == {}:
                    pass
                else:
                    if type(value) is CommissionRanges:
                        result[key] = self.converCommissionRanges(value)

                    # elif type(value) in [BakerStakeThreshold, ProtocolUpdate]:
                    #         result[key] = self.convertTypeWithSingleValues(value)

                    elif type(value) in self.simple_types:
                        result[key] = self.convertType(value)

                    elif type(value) in [CapitalBound, LeverageFactor]:
                        result[key] = self.convertTypeWithSingleValues(value)

                    elif type(value) is AmountFraction:
                        result[key] = self.convertType(value)

        return CCD_PoolParametersCpv1(**result)

    def convertElectionDifficulty(self, message) -> CCD_ElectionDifficulty:
        # TODO: no test available
        return message.value.parts_per_hundred_thousand / 100_000

    def convertMintRate(self, message) -> CCD_MintRate:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_MintRate(**result)

    def convertTimeParametersCpv1(self, message) -> CCD_TimeParametersCpv1:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_TimeParametersCpv1(**result)

    def convertCooldownParametersCpv1(self, message) -> CCD_CooldownParametersCpv1:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_CooldownParametersCpv1(**result)

    def convertConsensusParametersV1(self, message) -> CCD_ConsensusParametersV1:
        # TODO: no test available
        if MessageToDict(message) == {}:
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if self.valueIsEmpty(value):
                    pass
                else:
                    if type(value) is TimeoutParameters:
                        result[key] = self.convertTypeWithSingleValues(value)

                    elif type(value) in self.simple_types:
                        result[key] = self.convertType(value)
        # result = self.convertTypeWithSingleValues(message)

        return CCD_ConsensusParametersV1(**result)

    def convertFinalizationCommitteeParameters(
        self, message
    ) -> CCD_FinalizationCommitteeParameters:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_FinalizationCommitteeParameters(**result)

    def convertTransactionFeeDistribution(
        self, message
    ) -> CCD_TransactionFeeDistribution:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_TransactionFeeDistribution(**result)

    def convertGasRewards(self, message) -> CCD_GasRewards:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_GasRewards(**result)

    def convertGasRewardsV2(self, message) -> CCD_GasRewardsV2:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_GasRewardsV2(**result)

    def convertMintDistributionCpv1(self, message) -> CCD_MintDistributionCpv1:
        # TODO: no test available
        result = self.convertTypeWithSingleValues(message)

        return CCD_MintDistributionCpv1(**result)

    def convertMintDistributionCpv0(self, message) -> CCD_MintDistributionCpv0:
        # TODO: no test available
        if self.valueIsEmpty(message):
            return None
        else:
            result = {}
            for descriptor in message.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, message)
                if self.valueIsEmpty(value):
                    pass
                else:
                    if type(value) is MintRate:
                        result[key] = self.convertMintRate(value)

                    elif type(value) is AmountFraction:
                        result[key] = self.convertType(value)

        return CCD_MintDistributionCpv0(**result)

    def convertTokenModuleEvent(self, message) -> CCD_TokenModuleEvent:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if value == "":
                pass
            else:
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)

        return CCD_TokenModuleEvent(**result)

    def convertTokenTransferEvent(self, message) -> CCD_TokenTransferEvent:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif key == "amount":
                result[key] = CCD_TokenAmount(
                    **{"value": str(value.value), "decimals": value.decimals}
                )
        return CCD_TokenTransferEvent(**result)

    def convertTokenSupplyUpdateEvent(self, message) -> CCD_TokenSupplyUpdateEvent:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif key == "amount":
                result[key] = CCD_TokenAmount(
                    **{"value": str(value.value), "decimals": value.decimals}
                )
        return CCD_TokenSupplyUpdateEvent(**result)

    def convertDecimal(self, value: Decimal) -> CCD_TokenAmount:
        return CCD_TokenAmount(
            **{
                "value": "".join(str(digit) for digit in value.as_tuple().digits),
                "decimals": -1 * value.as_tuple().exponent,
            }
        )

    def convert_GovernanceAccount(self, cbor_decoded: dict) -> CCD_TokenHolder:
        return CCD_TokenHolder(
            account=self.convertAccountAddress(
                AccountAddress(value=cbor_decoded.get("governanceAccount").value[3])  # type: ignore
            )
        )

    def convertCreatePLT(self, message) -> CCD_CreatePLT:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if key == "initialization_parameters":
                cbor_decoded = cbor2.loads(value.value)
                if "initialSupply" in cbor_decoded:
                    token_amount = self.convertDecimal(
                        cbor_decoded.get("initialSupply")
                    )
                    cbor_decoded["initialSupply"] = token_amount
                cbor_decoded["governanceAccount"] = self.convert_GovernanceAccount(
                    cbor_decoded
                )
                if "metadata" in cbor_decoded:
                    # the value of key metadata is again a dict.
                    # we need to make sure all values in that dict are converted to str (can be bytes)
                    cbor_decoded["metadata"] = {
                        str(k): (
                            v
                            if isinstance(v, str)
                            else (
                                v.hex()
                                if isinstance(v, (bytes, bytearray, memoryview))
                                else str(v)
                            )
                        )
                        for k, v in cbor_decoded.get("metadata", {}).items()
                    }
                result[key] = CCD_InitializationParameters(**cbor_decoded)
            elif type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_CreatePLT(**result)

    def convertTokenEvents(self, message) -> list:
        events = []

        for entry in message:
            entry_dict: dict = {}
            for descriptor in entry.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, entry)
                if type(value) in self.simple_types:
                    converted_value = self.convertType(value)
                    if converted_value:
                        entry_dict[key] = converted_value
                elif MessageToDict(value) == {}:
                    pass
                elif type(value) is TokenModuleEvent:
                    entry_dict[key] = self.convertTokenModuleEvent(value)
                elif type(value) is TokenTransferEvent:
                    entry_dict[key] = self.convertTokenTransferEvent(value)
                elif type(value) is TokenSupplyUpdateEvent:
                    entry_dict[key] = self.convertTokenSupplyUpdateEvent(value)

            if entry_dict == {}:
                pass
            else:
                events.append(entry_dict)

        return events

    def convertEvents(self, message) -> list:
        events = []
        for entry in message:
            for descriptor in entry.DESCRIPTOR.fields:
                key, value = self.get_key_value_from_descriptor(descriptor, entry)
                if type(value) in self.simple_types:
                    converted_value = self.convertType(value)
                    if converted_value:
                        # entry_dict[key] = converted_value
                        events.append(converted_value)
            # if entry_dict == {}:
            #         pass
            # else:
            #     events.append(entry_dict)

        return events

    def convertInstanceInterruptedEvent(
        self, message
    ) -> CCD_ContractTraceElement_Interrupted:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            if key == "events":
                result[key] = self.convertEvents(value)

        return CCD_ContractTraceElement_Interrupted(**result)

    def convertInstanceUpdatedEvent(self, message) -> CCD_InstanceUpdatedEvent:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            if key == "events":
                result[key] = self.convertEvents(value)

        return CCD_InstanceUpdatedEvent(**result)

    def convertInstanceResumedEvent(self, message) -> CCD_ContractTraceElement_Resumed:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_ContractTraceElement_Resumed(**result)

    def convertInstanceTransferredEvent(
        self, message
    ) -> CCD_ContractTraceElement_Transferred:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_ContractTraceElement_Transferred(**result)

    def convertInstanceUpgradedEvent(
        self, message
    ) -> CCD_ContractTraceElement_Upgraded:
        result = {}
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_ContractTraceElement_Upgraded(**result)

    def convertUpdateEvents(self, message) -> list:
        events = []
        for entry in message:
            for descriptor in entry.DESCRIPTOR.fields:
                entry_dict = {}
                key, value = self.get_key_value_from_descriptor(descriptor, entry)
                if MessageToDict(value) == {}:
                    pass
                else:
                    if type(value) is InstanceUpdatedEvent:
                        entry_dict[key] = self.convertInstanceUpdatedEvent(value)

                    if type(value) is ContractTraceElement.Interrupted:
                        entry_dict[key] = self.convertInstanceInterruptedEvent(value)

                    if type(value) is ContractTraceElement.Resumed:
                        entry_dict[key] = self.convertInstanceResumedEvent(value)

                    if type(value) is ContractTraceElement.Transferred:
                        entry_dict[key] = self.convertInstanceTransferredEvent(value)

                    if type(value) is ContractTraceElement.Upgraded:
                        entry_dict[key] = self.convertInstanceUpgradedEvent(value)
                if entry_dict == {}:
                    pass
                else:
                    events.append(entry_dict)

        return events

    def convertCredentialRegistrationIdEntries(self, message):
        entries = []

        for list_entry in message:
            entries.append(self.convertType(list_entry))

        return entries

    def convertDuplicateCredIds(self, message) -> CCD_RejectReason_DuplicateCredIds:
        result = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            result[key] = self.convertCredentialRegistrationIdEntries(value)

        return CCD_RejectReason_DuplicateCredIds(**result)

    def convertRejectReason(self, message) -> CCD_RejectReason:
        result = {}
        _type = None
        for field, value in message.ListFields():
            key = field.name

            # Note this next section is purely to have Coverage
            # show us that we have not covered all possible reject
            # reasons with adequate tests...
            if key == "module_not_wf":
                test_me_please = True
            if key == "module_hash_already_exists":
                test_me_please = True
            if key == "invalid_account_reference":
                test_me_please = True
            if key == "invalid_init_method":
                test_me_please = True
            if key == "invalid_receive_method":
                test_me_please = True
            if key == "invalid_module_reference":
                test_me_please = True
            if key == "invalid_contract_address":
                test_me_please = True
            if key == "runtime_failure":
                test_me_please = True
            if key == "amount_too_large":
                test_me_please = True
            if key == "serialization_failure":
                test_me_please = True
            if key == "out_of_energy":
                test_me_please = True
            if key == "rejected_init":
                test_me_please = True
            if key == "rejected_receive":
                test_me_please = True
            if key == "invalid_proof":
                test_me_please = True
            if key == "already_a_baker: ":
                test_me_please = True
            if key == "not_a_baker":
                test_me_please = True
            if key == "insufficient_balance_for_baker_stake":
                test_me_please = True
            if key == "stake_under_minimum_threshold_for_baking":
                test_me_please = True
            if key == "baker_in_cooldown":
                test_me_please = True
            if key == "duplicate_aggregation_key":
                test_me_please = True
            if key == "non_existent_credential_id":
                test_me_please = True
            if key == "key_index_already_in_use":
                test_me_please = True
            if key == "invalid_account_threshold":
                test_me_please = True
            if key == "invalid_credential_key_sign_threshold":
                test_me_please = True
            if key == "invalid_encrypted_amount_transfer_proof":
                test_me_please = True
            if key == "invalid_transfer_to_public_proof":
                test_me_please = True
            if key == "encrypted_amount_self_transfer":
                test_me_please = True
            if key == "invalid_index_on_encrypted_transfer":
                test_me_please = True
            if key == "zero_scheduledAmount":
                test_me_please = True
            if key == "non_increasing_schedule":
                test_me_please = True
            if key == "first_scheduled_release_expired":
                test_me_please = True
            if key == "scheduled_self_transfer":
                test_me_please = True
            if key == "invalid_credentials":
                test_me_please = True
            if key == "duplicate_cred_ids":
                test_me_please = True
            if key == "non_existent_cred_ids":
                test_me_please = True
            if key == "remove_first_credential":
                test_me_please = True
            if key == "credential_holder_did_not_sign":
                test_me_please = True
            if key == "not_allowed_multiple_credentials":
                test_me_please = True
            if key == "not_allowed_to_receive_encrypted":
                test_me_please = True
            if key == "not_allowed_to_handle_encrypted":
                test_me_please = True
            if key == "missing_baker_add_parameters":
                test_me_please = True
            if key == "finalization_reward_commission_not_in_range":
                test_me_please = True
            if key == "baking_reward_commission_not_in_range":
                test_me_please = True
            if key == "transaction_fee_commission_not_in_range":
                test_me_please = True
            if key == "already_a_delegator":
                test_me_please = True
            if key == "insufficient_balance_for_delegation_stake":
                test_me_please = True
            if key == "missing_delegation_add_parameters":
                test_me_please = True
            if key == "insufficient_delegation_stak":
                test_me_please = True
            if key == "delegator_in_cooldown":
                test_me_please = True
            if key == "not_a_delegator":
                test_me_please = True
            if key == "delegation_target_not_a_baker: ":
                test_me_please = True
            if key == "stake_over_maximum_threshold_for_pool":
                test_me_please = True
            if key == "pool_would_become_over_delegated":
                test_me_please = True
            if key == "pool_closed":
                test_me_please = True  # noqa: F841

            _type = key
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) in [
                RejectReason.InvalidInitMethod,
                RejectReason.InvalidReceiveMethod,
                RejectReason.AmountTooLarge,
                RejectReason.RejectedInit,
                RejectReason.RejectedReceive,
            ]:
                result[key] = self.convertTypeWithSingleValues(value)

            elif type(value) is RejectReason.DuplicateCredIds:
                result[key] = self.convertDuplicateCredIds(value)
        return result, _type

    def convertRejectReasonNone(self, message) -> CCD_AccountTransactionEffects_None:
        result = {}
        _type = None
        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif type(value) is RejectReason:
                result[key], _type = self.convertRejectReason(value)

        return CCD_AccountTransactionEffects_None(**result), _type



# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import TokenInfoRequest, TokenInfo
from ccdexplorer_fundamentals.GRPCClient.protocol_level_tokens_pb2 import (
    TokenId,
    TokenState,
    TokenAmount,
)
from google.protobuf import message
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
import cbor2
import os
import sys
from ccdexplorer_fundamentals.enums import NET
from typing import TYPE_CHECKING


sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *

sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def convertTokenState(self, message: message.Message) -> CCD_TokenState:
        keys = {}

        for descriptor in message.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(descriptor, message)

            if key == "module_state":
                cbor_decoded = cbor2.loads(value.value)
                if "governanceAccount" in cbor_decoded:
                    cbor_decoded["governanceAccount"] = self.convert_GovernanceAccount(
                        cbor_decoded
                    )
                if "metadata" in cbor_decoded:
                    # the value of key metadata is again a dict.
                    # we need to make sure all values in that dict are converted to str (can be bytes)
                    cbor_decoded["metadata"] = {
                        str(k): (
                            v
                            if isinstance(v, str)
                            else (
                                v.hex()
                                if isinstance(v, (bytes, bytearray, memoryview))
                                else str(v)
                            )
                        )
                        for k, v in cbor_decoded.get("metadata", {}).items()
                    }
                keys[key] = CCD_ModuleState(**cbor_decoded)

            elif type(value) in self.simple_types:
                keys[key] = self.convertType(value)
            elif type(value) is TokenAmount:
                keys[key] = CCD_TokenAmount(
                    **{"value": str(value.value), "decimals": value.decimals}
                )

        return CCD_TokenState(**keys)

    def get_token_info(
        self: GRPCClient,  # type: ignore
        block_hash: str,
        token_id: str,
        net: Enum = NET.MAINNET,
    ) -> CCD_TokenInfo:
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        token_info_request = TokenInfoRequest(
            block_hash=blockHashInput, token_id=TokenId(value=token_id)
        )

        grpc_return_value: TokenInfo = self.stub_on_net(
            net, "GetTokenInfo", token_info_request
        )  # type: ignore

        result = {}
        for descriptor in grpc_return_value.DESCRIPTOR.fields:
            key, value = self.get_key_value_from_descriptor(
                descriptor, grpc_return_value
            )
            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

            elif isinstance(value, TokenState):
                result[key] = self.convertTokenState(value)

        return CCD_TokenInfo(**result)



# ruff: noqa :E501, F405
import datetime as dt
from enum import Enum
from typing import Dict, Optional, Union

import motor.motor_asyncio
from motor.motor_asyncio import AsyncIOMotorCollection
from pydantic import BaseModel, ConfigDict, Field
from pymongo import MongoClient, ReadPreference
from pymongo.collection import Collection
from rich.console import Console

from ccdexplorer_fundamentals.env import MONGO_URI
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *  # noqa: F403
from ccdexplorer_fundamentals.mongodb_queries._apy_calculations import (
    Mixin as _apy_calculations,
)
from ccdexplorer_fundamentals.mongodb_queries._baker_distributions import (
    Mixin as _distributions,
)
from ccdexplorer_fundamentals.mongodb_queries._search_transfers import (
    Mixin as _search_transfers,
)
from ccdexplorer_fundamentals.mongodb_queries._store_block import Mixin as _store_block
from ccdexplorer_fundamentals.mongodb_queries._subscriptions import (
    Mixin as _subscriptions,
)
from ccdexplorer_fundamentals.tooter import Tooter, TooterChannel, TooterType

console = Console()


class MongoLabeledAccount(BaseModel):
    """
    MongoLabeledAccount represents a labeled account in a MongoDB collection.

    Attributes:
        id (str): The unique identifier for the account, mapped from the MongoDB "_id" field.
        account_index (Optional[CCD_AccountIndex]): The index of the account, if available.
        label (str): The label associated with the account.
        label_group (str): The group to which the label belongs.
    """

    id: str = Field(..., alias="_id")
    account_index: Optional[CCD_AccountIndex] = None
    label: str
    label_group: str


class AccountStatementEntry(BaseModel):
    """
    AccountStatementEntry represents an entry in an account statement.

    Attributes:
        block_height (int): The height of the block in the blockchain.
        slot_time (dt.datetime): The timestamp of the slot.
        entry_type (str): The type of the entry (e.g., transaction, reward).
        amount (microCCD): The amount involved in the entry.
        balance (microCCD): The balance after the entry.
    """

    block_height: int
    slot_time: dt.datetime
    entry_type: str
    amount: microCCD
    balance: microCCD


class AccountStatementTransferType(BaseModel):
    """
    AccountStatementTransferType represents a transfer type in an account statement.

    Attributes:
        amount (microCCD): The amount of the transfer in microCCD.
        counterparty (str): The counterparty involved in the transfer.
    """

    amount: microCCD
    counterparty: str


class AccountStatementEntryType(BaseModel):
    """
    AccountStatementEntryType represents the structure of an account statement entry.

    Attributes:
        amount_decrypted (Optional[microCCD]): The decrypted amount in the account statement entry.
        amount_encrypted (Optional[microCCD]): The encrypted amount in the account statement entry.
        baker_reward (Optional[microCCD]): The reward received from baking.
        finalization_reward (Optional[microCCD]): The reward received from finalization.
        foundation_reward (Optional[microCCD]): The reward received from the foundation.
        transaction_fee (Optional[microCCD]): The transaction fee associated with the account statement entry.
        transaction_fee_reward (Optional[microCCD]): The reward received from transaction fees.
        transfer_in (Optional[list[AccountStatementTransferType]]): List of incoming transfers.
        transfer_out (Optional[list[AccountStatementTransferType]]): List of outgoing transfers.
    """

    amount_decrypted: Optional[microCCD] = None
    amount_encrypted: Optional[microCCD] = None
    baker_reward: Optional[microCCD] = None
    finalization_reward: Optional[microCCD] = None
    foundation_reward: Optional[microCCD] = None
    transaction_fee: Optional[microCCD] = None
    transaction_fee_reward: Optional[microCCD] = None
    transfer_in: Optional[list[AccountStatementTransferType]] = None
    transfer_out: Optional[list[AccountStatementTransferType]] = None


class MongoImpactedAddress(BaseModel):
    """
    MongoImpactedAddress is a data model representing an impacted address in a MongoDB collection.

    Attributes:
        id (str): The unique identifier for the impacted address, mapped from the MongoDB "_id" field.
        tx_hash (Optional[str]): The transaction hash associated with the impacted address. Rewards do not have a transaction hash.
        impacted_address (str): The impacted address.
        impacted_address_canonical (str): The canonical form of the impacted address.
        effect_type (str): The type of effect on the impacted address.
        balance_movement (Optional[AccountStatementEntryType]): The balance movement associated with the impacted address.
        block_height (int): The block height at which the impact occurred.
        included_in_flow (Optional[bool]): Indicates whether the impacted address is included in the flow.
        date (Optional[str]): The date when the impact occurred.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(..., alias="_id")
    tx_hash: Optional[str] = None  # Rewards do not have a tx
    impacted_address: str
    impacted_address_canonical: str
    effect_type: str
    balance_movement: Optional[AccountStatementEntryType] = None
    block_height: int
    included_in_flow: Optional[bool] = None
    date: Optional[str] = None
    plt_token_id: Optional[str] = None


class MongoTokensImpactedAddress(BaseModel):
    """
    MongoTokensImpactedAddress is a Pydantic model representing a MongoDB document for tokens impacted by an address.

    Attributes:
        id (str): The unique identifier for the document, mapped from MongoDB's "_id" field.
        tx_hash (str): The transaction hash associated with the impacted address.
        impacted_address (str): The address that was impacted.
        impacted_address_canonical (str): The canonical form of the impacted address.
        event_type (str): The type of event that impacted the address.
        token_address (Optional[str]): The address of the token, if applicable.
        plt_token_id (Optional[str]): The ID of the PLT token, if applicable.
        contract (Optional[str]): The contract associated with the token, if applicable.
        block_height (int): The block height at which the event occurred.
        date (str): The date when the event occurred.
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(..., alias="_id")
    tx_hash: str
    impacted_address: str
    impacted_address_canonical: str
    event_type: str
    token_address: Optional[str] = None
    plt_token_id: Optional[str] = None
    contract: Optional[str] = None
    block_height: int
    date: str


class MongoTypeBlockPerDay(BaseModel):
    """
    Block Per Day. This type is stored in the collection `blocks_per_day`.

    :Parameters:
    - `_id`: the date of the day that ended
    - `date`: the date of the day that ended
    - `height_for_first_block`: height of the first block in the day
    - `height_for_last_block`: height of the last block in the day
    - `slot_time_for_first_block`: time of the first block in the day
    - `slot_time_for_last_block`: time of the last block in the day
    - `hash_for_first_block`: hash of the first block in the day
    - `hash_for_last_block`: hash of the last block in the day

    """

    id: str = Field(..., alias="_id")
    date: str
    height_for_first_block: int
    height_for_last_block: int
    slot_time_for_first_block: dt.datetime
    slot_time_for_last_block: dt.datetime
    hash_for_first_block: str
    hash_for_last_block: str


class MongoTypeInvolvedAccount(BaseModel):
    """
    Involved Account. This type is stored in the collections `involved_accounts_all` and
    `involved_accounts_transfer`.

    :Parameters:
    - `_id`: the hash of the transaction
    - `sender`: the sender account address
    - `receiver`: the receiver account address, might be null
    - `sender_canonical`: the canonical sender account address
    - `receiver_canonical`: the  canonical receiver account address, might be null
    - `amount`: amount of the transaction, might be null
    - `type`: dict with transaction `type` and `contents`
    - `block_height`: height of the block in which the transaction is executed
    """

    id: str = Field(..., alias="_id")
    sender: str
    receiver: Optional[str] = None
    sender_canonical: str
    receiver_canonical: Optional[str] = None
    amount: Optional[int] = None
    type: dict[str, str]
    block_height: int
    memo: Optional[str] = None


class MongoTypeInvolvedContract(BaseModel):
    """
    Involved Contract. This type is stored in the collection `involved_contracts`.

    :Parameters:
    - `_id`: the hash of the transaction - the `str` representation of the contract.
    - `index`: contract index
    - `subindex`: contract subindex
    - `contract`: the `str` representation of the contract
    - `type`: dict with transaction `type` and `contents`
    - `block_height`: height of the block in which the transaction is executed
    - `source_module`: hash of the source module from which this contract is instanciated.
    """

    id: str = Field(..., alias="_id")
    index: int
    subindex: int
    contract: str
    type: dict[str, str]
    block_height: int
    source_module: str


class ModuleVerification(BaseModel):
    """
    ModuleVerification represents the verification details of a module.

    Attributes:
        verified (Optional[bool]): Indicates if the module is verified. Defaults to False.
        verification_status (str): The status of the verification process.
        verification_timestamp (Optional[dt.datetime]): The timestamp when the verification was performed. Defaults to None.
        explanation (Optional[str]): An optional explanation of the verification status. Defaults to None.
        build_image_used (Optional[str]): The build image used during verification. Defaults to None.
        build_command_used (Optional[str]): The build command used during verification. Defaults to None.
        archive_hash (Optional[str]): The hash of the archive used during verification. Defaults to None.
        link_to_source_code (Optional[str]): A link to the source code of the module. Defaults to None.
        source_code_at_verification_time (Optional[str]): The source code at the time of verification. Defaults to None.
    """

    verified: Optional[bool] = False
    verification_status: str
    verification_timestamp: Optional[dt.datetime] = None
    explanation: Optional[str] = None
    build_image_used: Optional[str] = None
    build_command_used: Optional[str] = None
    archive_hash: Optional[str] = None
    link_to_source_code: Optional[str] = None
    source_code_at_verification_time: Optional[str] = None


class MongoTypeModule(BaseModel):
    """
    MongoTypeModule represents a module stored in the `modules` collection.

    Attributes:
        id (str): The hex string identifier for the module, stored as `_id` in the database.
        module_name (str): The name of the module.
        methods (Optional[list[str]]): A list of method names associated with the module.
        contracts (Optional[list[str]]): A list of contract instances from this module.
        init_date (Optional[dt.datetime]): The initialization date of the module.
        verification (Optional[ModuleVerification]): The verification details of the module.
    """

    id: str = Field(..., alias="_id")
    module_name: Optional[str] = None
    methods: Optional[list[str]] = None
    contracts: Optional[list[str]] = None
    init_date: Optional[dt.datetime] = None
    verification: Optional[ModuleVerification] = None


class MongoTypeInstance(BaseModel):
    """
    Instance. This type is stored in the collection `instances`.

    :Parameters:
    - `_id`: the hex string
    - `module_name`: the name from the module
    - `methods`: list of method names
    - `contracts`: list of contract instances from this module
    """

    id: str = Field(..., alias="_id")
    v0: Optional[CCD_InstanceInfo_V0] = None  # noqa: F405
    v1: Optional[CCD_InstanceInfo_V1] = None  # noqa: F405
    source_module: Optional[str] = None
    module_verification: Optional[ModuleVerification] = None


class MongoTypeReward(BaseModel):
    """
    Module. This type is stored in the collection `payday_rewards`, property `reward`.

    """

    pool_owner: Optional[Union[int, str]] = None
    account_id: Optional[str] = None
    transaction_fees: int
    baker_reward: int
    finalization_reward: int


class MongoTypePoolReward(BaseModel):
    """
    Module. This type is stored in the collection `payday_rewards`.

    :Parameters:
    - `_id`: the hex string

    """

    id: str = Field(..., alias="_id")
    pool_owner: Union[int, str]
    pool_status: dict
    reward: MongoTypeReward
    date: str


class MongoTypeAccountReward(BaseModel):
    """
    Module. This type is stored in the collection `payday_rewards`.

    :Parameters:
    - `_id`: the hex string
    - `module_name`: the name from the module
    - `methods`: list of method names
    - `contracts`: list of contract instances from this module
    """

    id: str = Field(..., alias="_id")
    account_id: str
    staked_amount: int
    account_is_baker: Optional[bool] = None
    baker_id: Optional[int] = None
    reward: MongoTypeReward
    date: str


class Delegator(BaseModel):
    account: str
    stake: int


class MongoTypePayday(BaseModel):
    """
    Payday. This type is stored in collection `paydays`.

    :Parameters:
    - `_id`: hash of the block that contains payday information for
    this payday.
    - `date`: the payday date
    - `height_for_first_block`: height of the first block in the payday
    - `height_for_last_block`: height of the last block in the payday
    - `hash_for_first_block`: hash of the first block in the payday
    - `hash_for_last_block`: hash of the last block in the payday
    - `payday_duration_in_seconds`: duration of the payday in seconds (used for
    APY calculation)
    - `payday_block_slot_time`: time of payday reward block
    - `bakers_with_delegation_information`: bakers with delegators for reward period, retrieved
    from `get_delegators_for_pool_in_reward_period`, using the hash of the last block
    - `baker_account_ids`: mapping from baker_id to account_address
    - `pool_status_for_bakers`: dictionary, keyed on pool_status, value
    is a list of bakers, retrieved using the hash of the first block
    """

    id: str = Field(..., alias="_id")
    date: str
    height_for_first_block: int
    height_for_last_block: int
    hash_for_first_block: str
    hash_for_last_block: str
    payday_duration_in_seconds: float
    payday_block_slot_time: dt.datetime
    bakers_with_delegation_information: dict[str, list[Delegator]]
    baker_account_ids: dict[int, str]
    pool_status_for_bakers: Optional[dict[str, list[int]]] = None


class MongoTypePaydayV2(BaseModel):
    """
    Payday. This type is stored in collection `paydays_v2`.

    :Parameters:
    - `_id`: hash of the block that contains payday information for
    this payday.
    - `date`: the payday date
    - `height_for_first_block`: height of the first block in the payday
    - `height_for_last_block`: height of the last block in the payday
    - `hash_for_first_block`: hash of the first block in the payday
    - `hash_for_last_block`: hash of the last block in the payday
    - `payday_duration_in_seconds`: duration of the payday in seconds (used for
    APY calculation)
    - `payday_block_slot_time`: time of payday reward block
    """

    id: str = Field(..., alias="_id")
    date: str
    height_for_first_block: Optional[int] = None
    height_for_last_block: Optional[int] = None
    hash_for_first_block: Optional[str] = None
    hash_for_last_block: Optional[str] = None
    payday_duration_in_seconds: Optional[float] = None
    payday_block_slot_time: Optional[dt.datetime] = None


class MongoTypePaydayAPYIntermediate(BaseModel):
    """
    Payday APY Intermediate. This type is stored in collection `paydays_apy_intermediate`.

    :Parameters:
    - `_id`: baker_is or account address

    """

    id: str = Field(..., alias="_id")
    daily_apy_dict: dict
    d30_apy_dict: Optional[dict] = None
    d90_apy_dict: Optional[dict] = None
    d180_apy_dict: Optional[dict] = None


class MongoTypePaydaysPerformance(BaseModel):
    """
    Payday Performance. This is a collection that stores daily performance characteristics
    for bakers.


    :Parameters:
    - `_id`: unique id in the form of `date`-`baker_id`
    - `expectation`: the daily expected number of blocks for this baker in this payday.
    Calculated as the lottery power * 8640 (the expected number of blocks in a day)
    - `payday_block_slot_time`: Slot time of the payday block
    - `baker_id`: the baker_id
    - `pool_status`:

    """

    id: str = Field(..., alias="_id")
    pool_status: CCD_PoolInfo  # noqa: F405
    expectation: float
    date: str
    payday_block_slot_time: dt.datetime
    baker_id: str


class Collections(Enum):
    """
    Enum class representing various MongoDB collection names used in the application.

    Attributes:
        all_account_addresses: Collection for all account addresses.
        blocks: Collection for blocks.
        blocks_log: Collection for blocks log.
        transactions: Collection for transactions.
        special_events: Collection for special events.
        instances: Collection for instances.
        modules: Collection for modules.
        paydays: Collection for paydays.
        paydays_performance: Collection for paydays performance.
        paydays_rewards: Collection for paydays rewards.
        paydays_apy_intermediate: Collection for paydays APY intermediate.
        paydays_current_payday: Collection for current payday.
        paydays_helpers: Collection for payday helpers.
        involved_accounts_transfer: Collection for involved accounts transfer.
        nightly_accounts: Collection for nightly accounts.
        blocks_at_end_of_day: Collection for blocks at end of day.
        blocks_per_day: Collection for blocks per day.
        helpers: Collection for helpers.
        memo_transaction_hashes: Collection for memo transaction hashes.
        cns_domains: Collection for CNS domains.
        dashboard_nodes: Collection for dashboard nodes.
        tokens_accounts: Collection for token accounts.
        tokens_links_v2: Collection for token links version 2.
        tokens_links_v3: Collection for token links version 3.
        tokens_token_addresses_v2: Collection for token addresses version 2.
        tokens_tags: Collection for token tags.
        tokens_logged_events: Collection for token logged events.
        tokens_logged_events_v2: Collection for token logged events version 2.
        tokens_token_addresses: Collection for token addresses.
        memos_to_hashes: Collection for memos to hashes.
        credentials_issuers: Collection for credential issuers.
        impacted_addresses: Collection for impacted addresses.
        impacted_addresses_pre_payday: Collection for impacted addresses pre-payday.
        impacted_addresses_all_top_list: Collection for all top list of impacted addresses.
        pre_tokens_overview: Collection for pre-rendered tokens overview.
        pre_addresses_by_contract_count: Collection for pre-rendered addresses by contract count.
        pre_tokens_by_address: Collection for pre-rendered tokens by address.
        statistics: Collection for statistics.
        pre_render: Collection for pre-rendered data.
        cis5_public_keys_contracts: Collection for CIS5 public keys contracts.
        cis5_public_keys_info: Collection for CIS5 public keys info.
        projects: Collection for projects.
        usecases: Collection for use cases.
        tokens_impacted_addresses: Collection for tokens impacted addresses.
        tnt_logged_events: Collection for TNT logged events.
        queue_todo: Collection for queue to-do items.
        tx_types_count: Collection for transaction types count.
    """

    all_account_addresses = "all_account_addresses"
    blocks = "blocks"
    blocks_log = "blocks_log"
    transactions = "transactions"
    special_events = "special_events"
    instances = "instances"
    modules = "modules"
    paydays = "paydays"
    paydays_performance = "paydays_performance"
    paydays_rewards = "paydays_rewards"
    paydays_apy_intermediate = "paydays_apy_intermediate"
    paydays_current_payday = "paydays_current_payday"
    paydays_v2 = "paydays_v2"
    paydays_v2_performance = "paydays_v2_performance"
    paydays_v2_rewards = "paydays_v2_rewards"
    paydays_v2_apy = "paydays_v2_apy"
    paydays_v2_current_payday = "paydays_v2_current_payday"
    paydays_helpers = "paydays_helpers"
    involved_accounts_transfer = "involved_accounts_transfer"
    nightly_accounts = "nightly_accounts"
    blocks_at_end_of_day = "blocks_at_end_of_day"
    blocks_per_day = "blocks_per_day"
    helpers = "helpers"
    memo_transaction_hashes = "memo_transaction_hashes"
    cns_domains = "cns_domains"
    dashboard_nodes = "dashboard_nodes"
    tokens_accounts = "tokens_accounts"
    tokens_links_v2 = "tokens_links_v2"
    tokens_links_v3 = "tokens_links_v3"
    tokens_token_addresses_v2 = "tokens_token_addresses_v2"
    tokens_tags = "tokens_tags"
    tokens_logged_events = "tokens_logged_events"
    tokens_logged_events_v2 = "tokens_logged_events_v2"
    tokens_token_addresses = "tokens_token_addresses"
    memos_to_hashes = "memos_to_hashes"
    credentials_issuers = "credentials_issuers"
    impacted_addresses = "impacted_addresses"
    impacted_addresses_pre_payday = "impacted_addresses_pre_payday"
    impacted_addresses_all_top_list = "impacted_addresses_all_top_list"
    # statistics and pre-renders
    pre_tokens_overview = "pre_tokens_overview"
    pre_addresses_by_contract_count = "pre_addresses_by_contract_count"
    pre_tokens_by_address = "pre_tokens_by_address"
    statistics = "statistics"
    pre_render = "pre_render"
    cis5_public_keys_contracts = "cis5_public_keys_contracts"
    cis5_public_keys_info = "cis5_public_keys_info"
    tx_types_count = "tx_types_count"
    # addresses and contracts per net per usecase
    projects = "projects"
    usecases = "usecases"
    tokens_impacted_addresses = "tokens_impacted_addresses"
    tnt_logged_events = "tnt_logged_events"
    queue_todo = "queue_todo"
    validator_logs = "validator_logs"
    blocks_with_only_chain_txs = "blocks_with_only_chain_txs"
    # PLT
    plts_links = "plts_links"
    plts_tags = "plts_tags"


class CollectionsUtilities(Enum):
    """
    Enum class representing various MongoDB collection names used in the application.

    Attributes:
        labeled_accounts (str): Collection name for labeled accounts.
        labeled_accounts_metadata (str): Collection name for labeled accounts metadata.
        exchange_rates (str): Collection name for exchange rates.
        exchange_rates_historical (str): Collection name for historical exchange rates.
        users_v2_prod (str): Collection name for production users (version 2).
        users_v2_dev (str): Collection name for development users (version 2).
        message_log (str): Collection name for message logs.
        preferences_explanations (str): Collection name for preferences explanations.
        release_notes (str): Collection name for release notes.
        token_api_translations (str): Collection name for token API translations.
        projects (str): Collection name for projects (use case management).
        usecases (str): Collection name for use cases (use case management).
        helpers (str): Collection name for helpers (use case management).
        api_api_keys (str): Collection name for API keys.
        api_users (str): Collection name for API users.
    """

    labeled_accounts = "labeled_accounts"
    labeled_accounts_metadata = "labeled_accounts_metadata"
    exchange_rates = "exchange_rates"
    exchange_rates_historical = "exchange_rates_historical"
    users_v2_prod = "users_v2_prod"
    users_v2_dev = "users_v2_dev"
    message_log = "message_log"
    preferences_explanations = "preferences_explanations"
    release_notes = "release_notes"
    token_api_translations = "token_api_translations"
    # use case management
    projects = "projects"
    usecases = "usecases"
    helpers = "helpers"
    # api
    api_api_keys = "api_api_keys"
    # api
    api_users = "api_users"


class MongoDB(
    _search_transfers,
    _subscriptions,
    _distributions,
    _store_block,
    _apy_calculations,
):
    def __init__(self, tooter: Tooter, nearest: bool = False):
        self.tooter: Tooter = tooter
        try:
            if nearest:
                con = MongoClient(MONGO_URI, read_preference=ReadPreference.NEAREST)
            else:
                con = MongoClient(MONGO_URI)
            self.connection: MongoClient = con

            self.mainnet_db = con["concordium_mainnet"]
            self.mainnet: Dict[Collections, Collection] = {}
            for collection in Collections:
                self.mainnet[collection] = self.mainnet_db[collection.value]

            self.testnet_db = con["concordium_testnet"]
            self.testnet: Dict[Collections, Collection] = {}
            for collection in Collections:
                self.testnet[collection] = self.testnet_db[collection.value]

            self.devnet_db = con["concordium_devnet"]
            self.devnet: Dict[Collections, Collection] = {}
            for collection in Collections:
                self.devnet[collection] = self.devnet_db[collection.value]

            self.utilities_db = con["concordium_utilities"]
            self.utilities: Dict[CollectionsUtilities, Collection] = {}
            for collection in CollectionsUtilities:
                self.utilities[collection] = self.utilities_db[collection.value]

            console.log(con.server_info()["version"])
        except Exception as e:
            print(e)
            tooter.send(
                channel=TooterChannel.NOTIFIER,
                message=f"BOT ERROR! Cannot connect to MongoDB, with error: {e}",
                notifier_type=TooterType.MONGODB_ERROR,
            )


class MongoMotor(
    _search_transfers,
    _subscriptions,
    _distributions,
    _store_block,
    _apy_calculations,
):
    def __init__(self, tooter: Tooter, nearest: bool = False):
        self.tooter: Tooter = tooter
        try:
            if nearest:
                con = motor.motor_asyncio.AsyncIOMotorClient(
                    MONGO_URI, read_preference=ReadPreference.NEAREST
                )
            else:
                con = motor.motor_asyncio.AsyncIOMotorClient(MONGO_URI)
            self.connection = con

            self.mainnet_db = con["concordium_mainnet"]
            self.mainnet: Dict[Collections, AsyncIOMotorCollection] = {}  # type: ignore
            for collection in Collections:
                self.mainnet[collection] = self.mainnet_db[collection.value]

            self.testnet_db = con["concordium_testnet"]
            self.testnet: Dict[Collections, AsyncIOMotorCollection] = {}  # type: ignore
            for collection in Collections:
                self.testnet[collection] = self.testnet_db[collection.value]

            self.utilities_db = con["concordium_utilities"]
            self.utilities: Dict[CollectionsUtilities, AsyncIOMotorCollection] = {}  # type: ignore
            for collection in CollectionsUtilities:
                self.utilities[collection] = self.utilities_db[collection.value]
            # console.log(f'Motor: {con.server_info()["version"]}')
        except Exception as e:
            print(e)
            tooter.send(
                channel=TooterChannel.NOTIFIER,
                message=f"BOT ERROR! Cannot connect to Motor MongoDB, with error: {e}",
                notifier_type=TooterType.MONGODB_ERROR,
            )



# # ruff: noqa: F403, F405, E402
# from __future__ import annotations
# from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
# from ccdexplorer_fundamentals.enums import NET
# from enum import Enum
# from typing import TYPE_CHECKING

# if TYPE_CHECKING:
#     from ccdexplorer_fundamentals.GRPCClient import GRPCClient
# from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
#     Mixin as _SharedConverters,
# )
# import os
# import sys

# sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
# from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *


# class Mixin(_SharedConverters):
#     def convertSuccess(self, message) -> CCD_InvokeInstanceResponse_Success:
#         result = {}
#         for descriptor in message.DESCRIPTOR.fields:
#             key, value = self.get_key_value_from_descriptor(descriptor, message)
#             if key == "effects":
#                 result[key] = self.convertUpdateEvents(value)

#             elif type(value) in self.simple_types:
#                 result[key] = self.convertType(value)

#         return CCD_InvokeInstanceResponse_Success(**result)

#     def convertFailure(self, message) -> CCD_InvokeInstanceResponse_Failure:
#         result = {}
#         for descriptor in message.DESCRIPTOR.fields:
#             key, value = self.get_key_value_from_descriptor(descriptor, message)
#             if type(value) is RejectReason:
#                 result[key], _ = self.convertRejectReason(value)

#             elif type(value) in self.simple_types:
#                 result[key] = self.convertType(value)

#         return CCD_InvokeInstanceResponse_Failure(**result)

#     def send_account_transaction(
#         self: GRPCClient,
#         block_hash: str,
#         contract_name: str,
#         instance_index: int,
#         instance_subindex: int,
#         entrypoint: str,
#         parameter_bytes: bytes,
#         net: Enum = NET.MAINNET,
#     ) -> TransactionHash:
#         result = {}
#         pass
#         # signature = AccountTransactionSignature()
#         # update_header = AccountTransactionHeader(
#         #     expiry: TransactionExpiry.futureMinutes(60),
#         #     nonce: (await client.getNextAccountNonce(sender)).nonce,
#         #     sender,
#         #     )

#         # update_params = serializeUpdateContractParameters(
#         #     contract_name,
#         #     entrypoint,
#         #     rainyWeather,
#         #     schema
#         # )
#         # update_payload: UpdateContractPayload = {
#         #     amount: CcdAmount.zero(),
#         #     address: unwrap(contractAddress),
#         #     receiveName,
#         #     message: update_params,
#         #     maxContractExecutionEnergy: maxCost,
#         # }
#         # payload = AccountTransactionPayload()
#         # update_signature =  signTransaction(updateTransaction, signer);
#         # update_transaction= AccountTransaction(
#         #     signature=update_signature,
#         #     header= update_header,
#         #     payload= update_payload
#         # )

#         # updateTrxHash = sendAccountTransaction(
#         #     updateTransaction,
#         #     updateSignature
#         # );

#         # blockHashInput = self.generate_block_hash_input_from(block_hash)
#         # invokeInstanceRequest = self.generate_invoke_instance_request_from(
#         #     instance_index,
#         #     instance_subindex,
#         #     blockHashInput,
#         #     entrypoint,
#         #     parameter_bytes,
#         # )

#         # grpc_return_value: InvokeInstanceResponse = self.stub_on_net(
#         #     net, "InvokeInstance", invokeInstanceRequest
#         # )

#         # for descriptor in grpc_return_value.DESCRIPTOR.fields:
#         #     key, value = self.get_key_value_from_descriptor(
#         #         descriptor, grpc_return_value
#         #     )

#         #     if type(value) is InvokeInstanceResponse.Success:
#         #         result[key] = self.convertSuccess(value)

#         #     elif type(value) is InvokeInstanceResponse.Failure:
#         #         result[key] = self.convertFailure(value)

#         # return CCD_InvokeInstanceResponse(**result)



# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.GRPCClient.protocol_level_tokens_pb2 import (
    TokenId,
)
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *
from ccdexplorer_fundamentals.enums import NET
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def get_token_list(
        self: GRPCClient,  # type: ignore
        block_hash: str,
        net: Enum = NET.MAINNET,
    ) -> list[CCD_TokenId]:
        result = []
        blockHashInput = self.generate_block_hash_input_from(block_hash)

        grpc_return_value: list[TokenId] = self.stub_on_net(
            net, "GetTokenList", blockHashInput
        )  # type: ignore

        for token in list(grpc_return_value):
            result.append(self.convertType(token))

        return result



from enum import Enum
import dateutil
from .cns import CNSDomain, CNSActions

from ccdexplorer_fundamentals.block import ConcordiumBlockInfo
import io
from rich.console import Console

console = Console()


# sender, receiver, txHash, contract, amount, tokenId, self.cns_domain.domain_name, self
class ClassificationResult:
    def __init__(self):
        self.sender = None
        self.receiver = None
        self.txHash = None
        self.contract = None
        self.amount = None
        self.tokenId = None
        self.domain_name = None
        self.type = None
        self.contents = None
        self.accounts_involved_all = False
        self.accounts_involved_transfer = False
        self.contracts_involved = False
        self.list_of_contracts_involved = set()


class DelegationAndStakingResult:
    def __init__(self):
        self.sender = None
        self.bakerId = None
        self.txHash = None
        self.message = ""
        self.request_target = None
        self.perc = None
        self.unstaked = False
        self.unstaked_amount = 0


class TransactionClass(Enum):
    AccountTransaction = "accountTransaction"
    CredentialDeploymentTransaction = "credentialDeploymentTransaction"
    UpdateTransaction = "updateTransaction"


class TransactionType(Enum):
    # https://github.com/Concordium/concordium-scan/blob/e95e8b2b191fefcf381ef4b4a1c918dd1f11ae05/frontend/src/queries/useTransactionQuery.ts
    AccountCreated = """
    ... on AccountCreated {
        accountAddress {
            asString
        }
    }
    """
    AmountAddedByDecryption = """
    ... on AmountAddedByDecryption {
        amount
        accountAddress {
            asString
        }
    }
    """
    BakerAdded = """
    ... on BakerAdded {
        bakerId
        restakeEarnings
        stakedAmount
        electionKey
        signKey
        aggregationKey
        accountAddress {
            asString
        }
    }
    """
    BakerKeysUpdated = """
    ... on BakerKeysUpdated {
        bakerId
        electionKey
        signKey
        aggregationKey
        accountAddress {
            asString
        }
    }
    """
    BakerRemoved = """
    ... on BakerRemoved {
        bakerId
        accountAddress {
            asString
        }
    }
    """
    BakerSetBakingRewardCommission = """
    ... on BakerSetBakingRewardCommission {
        bakerId
        bakingRewardCommission
        accountAddress {
            asString
        }
    }
    """
    BakerSetFinalizationRewardCommission = """
    ... on BakerSetFinalizationRewardCommission {
        bakerId
        finalizationRewardCommission
        accountAddress {
            asString
        }
    }
    """
    BakerSetMetadataURL = """
    ... on BakerSetMetadataURL {
        bakerId
        metadataUrl
        accountAddress {
            asString
        }
    }
    """
    BakerSetOpenStatus = """
    ... on BakerSetOpenStatus {
        bakerId
        openStatus
        accountAddress {
            asString
        }
    }
    """
    BakerSetRestakeEarnings = """
    ... on BakerSetRestakeEarnings {
        bakerId
        restakeEarnings
        accountAddress {
            asString
        }
    }    
    """
    BakerSetTransactionFeeCommission = """
    ... on BakerSetTransactionFeeCommission {
        bakerId
        transactionFeeCommission
        accountAddress {
            asString
        }
    }
    """
    BakerStakeDecreased = """
    ... on BakerStakeDecreased {
        bakerId
        newStakedAmount
        accountAddress {
            asString
        }
    }
    """
    BakerStakeIncreased = """
    ... on BakerStakeIncreased {
        bakerId
        newStakedAmount
        accountAddress {
            asString
        }
    }
    """
    ChainUpdateEnqueued = """
    ... on ChainUpdateEnqueued {
        effectiveImmediately
        __typename
	effectiveTime
	payload {
		__typename
		...on AddAnonymityRevokerChainUpdatePayload {
			name
		}
		...on AddIdentityProviderChainUpdatePayload {
			name
		}
		...on BakerStakeThresholdChainUpdatePayload {
			amount
		}
		...on ElectionDifficultyChainUpdatePayload {
			electionDifficulty
		}
		...on EuroPerEnergyChainUpdatePayload {
			exchangeRate {
				numerator
				denominator
			}
		}
		...on FoundationAccountChainUpdatePayload {
			accountAddress {
				asString
			}
		}
		...on GasRewardsChainUpdatePayload {
			accountCreation
			baker
			chainUpdate
			finalizationProof
		}
		...on MicroCcdPerEuroChainUpdatePayload {
			exchangeRate {
				denominator
				numerator
			}
		}
		...on MintDistributionChainUpdatePayload {
			bakingReward
			finalizationReward
			mintPerSlot
		}
		...on ProtocolChainUpdatePayload {
			message
			specificationUrl
		}
		...on TransactionFeeDistributionChainUpdatePayload {
			baker
			gasAccount
		}
		...on CooldownParametersChainUpdatePayload {
			delegatorCooldown
			poolOwnerCooldown
		}
		...on TimeParametersChainUpdatePayload {
			mintPerPayday
			rewardPeriodLength
		}
		...on MintDistributionV1ChainUpdatePayload {
			bakingReward
			finalizationReward
		}
		...on PoolParametersChainUpdatePayload {
			bakingCommissionRange {
				min
				max
			}
			finalizationCommissionRange {
				min
				max
			}
			transactionCommissionRange {
				min
				max
			}
			passiveBakingCommission
			passiveFinalizationCommission
			passiveTransactionCommission
			minimumEquityCapital
			capitalBound
			leverageBound {
				denominator
				numerator
			}
		}
	}
    }
    """
    ContractInitialized = """
    ... on ContractInitialized {
        amount
        contractAddress {
            __typename
            asString
            index
            subIndex
        }
        HEX:eventsAsHex {
            nodes 
        }
        initName
        moduleRef
    }
    """
    ContractInterrupted = """
    ... on ContractInterrupted {
        contractAddress {
            __typename
            asString
            index
            subIndex
        }
        hex_events:eventsAsHex
    }
    """
    ContractModuleDeployed = """
    ... on ContractModuleDeployed {
        moduleRef
        __typename
    }
    """
    ContractResumed = """
    ... on ContractResumed {
        contractAddress {
            __typename
            asString
            index
            subIndex
        }
        success
    }
    """
    ContractUpdated = """
    ... on ContractUpdated {
        amount
        contractAddress {
            __typename
            asString
            index
            subIndex
        }
        eventsAsHex {
            nodes 
        }
        instigator {
            __typename
            ... on AccountAddress {
                __typename
                asString
            }
            ... on ContractAddress {
                __typename
                asString
                index
                subIndex
            }
        }
        messageAsHex
        receiveName
    }
    """
    CredentialDeployed = """
    ... on CredentialDeployed {
        accountAddress {
            asString
        }
        regId
    }
    """
    CredentialKeysUpdated = """
    ... on CredentialKeysUpdated {
        credId
    }
    """
    CredentialsUpdated = """
    ... on CredentialsUpdated {
        accountAddress {
            asString
        }
        newCredIds
        newThreshold
        removedCredIds
    }
    """
    DataRegistered = """
    ... on DataRegistered {
        dataAsHex
    }
    """
    DelegationAdded = """
    ... on DelegationAdded {
        accountAddress {
            asString
        }
        delegatorId
    }
    """
    DelegationRemoved = """
    ... on DelegationRemoved {
        accountAddress {
            asString
        }
        delegatorId
    }
    """
    DelegationSetDelegationTarget = """
    ... on DelegationSetDelegationTarget {
        accountAddress {
            asString
        }
        delegationTarget {
            __typename
            ... on BakerDelegationTarget {
                bakerId
            }
        }
        delegatorId
    }
    """
    DelegationSetRestakeEarnings = """
    ... on DelegationSetRestakeEarnings {
        accountAddress {
            asString
        }
        delegatorId
        restakeEarnings
    }
    """
    DelegationStakeDecreased = """
    ... on DelegationStakeDecreased {
        accountAddress {
            asString
        }
        delegatorId
        newStakedAmount
    }
    """
    DelegationStakeIncreased = """
    ... on DelegationStakeIncreased {
        accountAddress {
            asString
        }
        delegatorId
        newStakedAmount
    }
    """
    EncryptedAmountsRemoved = """
    ... on EncryptedAmountsRemoved {
        accountAddress {
                asString
            }
        inputAmount
        newEncryptedAmount
        upToIndex
    }
    """
    EncryptedSelfAmountAdded = """
    ... on EncryptedSelfAmountAdded {
        accountAddress {
                asString
            }
        amount
        newEncryptedAmount
    }
    """
    NewEncryptedAmount = """
    ... on NewEncryptedAmount {
        accountAddress {
                asString
            }
        encryptedAmount
        newIndex
    }
    """
    TransferMemo = """
    ... on TransferMemo {
        decoded {
            text
        }
        rawHex
    }
    """
    Transferred = """
    ... on Transferred {
        amount
        from {
            ... on AccountAddress {
            __typename
            asString
            }
            ... on ContractAddress {
            __typename
            index
            subIndex
            }
            __typename
        }
        to {
            ... on AccountAddress {
            __typename
            asString
            }
            ... on ContractAddress {
            __typename
            index
            subIndex
            asString
            }
            __typename
        }
        __typename
        }
    """
    TransferredWithSchedule = """
    ... on TransferredWithSchedule {
        amountsSchedule (first: 50) {
            nodes {
                amount
                timestamp
            }
        }
        fromAccountAddress {
            asString
        }
            
        toAccountAddress {
            asString
        }
        totalAmount
        }
        
    """


class EventOpenStatusFromQLToNode(Enum):
    OPEN_FOR_ALL = "openForAll"
    CLOSED_FOR_NEW = "closedForNew"
    CLOSED_FOR_ALL = "closedForAll"


class TransactionTypeFromQLToNode(Enum):
    AccountTransaction = "accountTransaction"
    CredentialDeploymentTransaction = "credentialDeploymentTransaction"
    UpdateTransaction = "updateTransaction"


class TransactionContentsFromQLToNode(Enum):
    # AccountTransaction
    ADD_BAKER = "addBaker"
    CONFIGURE_BAKER = "configureBaker"
    CONFIGURE_DELEGATION = "configureDelegation"
    DEPLOY_MODULE = "deployModule"
    ENCRYPTED_TRANSFER = "encryptedAmountTransfer"
    ENCRYPTED_TRANSFER_WITH_MEMO = "encryptedAmountTransferWithMemo"
    INITIALIZE_SMART_CONTRACT_INSTANCE = "initContract"
    REGISTER_DATA = "registerData"
    REMOVE_BAKER = "removeBaker"
    SIMPLE_TRANSFER = "transfer"
    SIMPLE_TRANSFER_WITH_MEMO = "transferWithMemo"
    TRANSFER_TO_ENCRYPTED = "transferToEncrypted"
    TRANSFER_TO_PUBLIC = "transferToPublic"
    TRANSFER_WITH_SCHEDULE = "transferWithSchedule"
    TRANSFER_WITH_SCHEDULE_WITH_MEMO = "transferWithScheduleAndMemo"
    UPDATE_BAKER_KEYS = "updateBakerKeys"
    UPDATE_BAKER_RESTAKE_EARNINGS = "updateBakerRestakeEarnings"
    UPDATE_BAKER_STAKE = "updateBakerStake"
    UPDATE_CREDENTIAL_KEYS = "updateCredentialKeys"
    UPDATE_CREDENTIALS = "updateCredentials"
    UPDATE_SMART_CONTRACT_INSTANCE = "update"
    UNKNOWN = "_unknown"

    # CredentialDeploymentTransaction
    NORMAL = "normal"
    INITIAL = "initial"

    # UpdateTransaction
    UPDATE_ADD_ANONYMITY_REVOKER = "updateAddAnonymityRevoker"
    UPDATE_ADD_IDENTITY_PROVIDER = "updateAddIdentityProvider"
    UPDATE_BAKER_STAKE_THRESHOLD = "updateBakerStake"
    UPDATE_COOLDOWN_PARAMETERS = "updateCooldownParameters"
    UPDATE_ELECTION_DIFFICULTY = "updateElectionDifficulty"
    UPDATE_EURO_PER_ENERGY = "updateEuroPerEnergy"
    UPDATE_FOUNDATION_ACCOUNT = "updateFoundationAccount"
    UPDATE_GAS_REWARDS = "updateGASRewards"
    UPDATE_LEVEL1_KEYS = "updateLevel1Keys"
    UPDATE_LEVEL2_KEYS = "updateLevel2Keys"
    UPDATE_MICRO_GTU_PER_EURO = "updateMicroGTUPerEuro"
    UPDATE_MINT_DISTRIBUTION = "updateMintDistribution"
    UPDATE_POOL_PARAMETERS = "updatePoolParameters"
    UPDATE_PROTOCOL = "updateProtocol"
    UPDATE_ROOT_KEYS = "updateRootKeys"
    UPDATE_TIME_PARAMETERS = "updateTimeParameters"
    UPDATE_TRANSACTION_FEE_DISTRIBUTION = "updateTransactionFeeDistribution"


class Reward:
    """ """

    def __init__(self, r):
        self.r = r

    def account_reward(self):
        r = self.r
        self.tag = r["tag"]

        if self.tag == "PaydayAccountReward":
            return r["account"]
        else:
            return None

    def pool_reward(self):
        r = self.r
        self.tag = r["tag"]

        if self.tag == "PaydayPoolReward":
            return r["poolOwner"]
        else:
            return None


class Event:
    def __init__(self, _event):
        self._event = _event
        self.d = {}

    def determineAddress(self, address):
        a = {}

        if address["__typename"] == "AccountAddress":
            a = {"type": "AddressAccount", "address": address["asString"]}
        elif address["__typename"] == "ContractAddress":
            a = {
                "type": "AddressContract",
                "address": {"index": address["index"], "subindex": address["subIndex"]},
            }
        return a

    def decode_memo(self, hex):
        # bs = bytes.fromhex(hex)
        # return bytes.decode(bs[1:], 'UTF-8')
        try:
            bs = io.BytesIO(bytes.fromhex(hex))
            n = int.from_bytes(bs.read(1), byteorder="little")
            value = bs.read(n)
            try:
                memo = bytes.decode(value, "UTF-8")
                return memo
            except UnicodeDecodeError:
                memo = bytes.decode(value[1:], "UTF-8")
                return memo
        except:
            return "Decoding failure..."

    def translate_memo_if_present(self):
        if self._event["tag"] == "TransferMemo":
            self._event["memo"] = self.decode_memo(self._event["memo"])

    def translate_event_from_graphQL(self):
        e = self._event
        etype = e["__typename"]
        self.d["tag"] = etype

        if etype == TransactionType.AccountCreated.name:
            self.d["contents"] = e["accountAddress"]["asString"]

        if etype == TransactionType.AmountAddedByDecryption.name:
            self.d["amount"] = int(e["amount"])
            self.d["account"] = e["accountAddress"]["asString"]

        elif etype == TransactionType.BakerSetOpenStatus.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["openStatus"] = EventOpenStatusFromQLToNode[e["openStatus"]].value

        elif etype == TransactionType.BakerSetBakingRewardCommission.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["bakingRewardCommission"] = e["bakingRewardCommission"]

        elif etype == TransactionType.BakerSetFinalizationRewardCommission.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["finalizationRewardCommission"] = e["finalizationRewardCommission"]

        elif etype == TransactionType.BakerSetTransactionFeeCommission.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["transactionFeeCommission"] = e["transactionFeeCommission"]

        elif etype == TransactionType.BakerSetMetadataURL.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["metadataUrl"] = e["metadataUrl"]

        elif etype == TransactionType.DelegationAdded.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["delegatorId"] = e["delegatorId"]

        elif etype == TransactionType.DelegationSetDelegationTarget.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["delegatorId"] = e["delegatorId"]
            if e["delegationTarget"]["__typename"] == "PassiveDelegationTarget":
                self.d["delegationTarget"] = {"delegateType": "Passive"}
            else:
                self.d["delegationTarget"] = {
                    "delegateType": "Baker",
                    "bakerId": e["delegationTarget"]["bakerId"],
                }

        elif etype == TransactionType.DelegationSetRestakeEarnings.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["delegatorId"] = e["delegatorId"]
            self.d["restakeEarnings"] = e["restakeEarnings"]

        elif etype == TransactionType.DelegationRemoved.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["delegatorId"] = e["delegatorId"]

        elif etype == TransactionType.DelegationStakeDecreased.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["delegatorId"] = e["delegatorId"]
            self.d["newStake"] = e["newStakedAmount"]

        elif etype == TransactionType.DelegationStakeIncreased.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["delegatorId"] = e["delegatorId"]
            self.d["newStake"] = e["newStakedAmount"]

        elif etype == TransactionType.BakerAdded.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["aggregationKey"] = e["aggregationKey"]
            self.d["electionKey"] = e["electionKey"]
            self.d["restakeEarnings"] = e["restakeEarnings"]
            self.d["signKey"] = e["signKey"]
            self.d["stake"] = e["stakedAmount"]

        elif etype == TransactionType.BakerKeysUpdated.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["aggregationKey"] = e["aggregationKey"]
            self.d["electionKey"] = e["electionKey"]
            self.d["signKey"] = e["signKey"]

        elif etype == TransactionType.BakerRemoved.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]

        elif etype == TransactionType.BakerSetRestakeEarnings.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["restakeEarnings"] = e["restakeEarnings"]

        elif etype == TransactionType.BakerStakeDecreased.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["newStake"] = e["newStakedAmount"]

        elif etype == TransactionType.BakerStakeIncreased.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["bakerId"] = e["bakerId"]
            self.d["newStake"] = e["newStakedAmount"]

        elif etype == TransactionType.ChainUpdateEnqueued.name:
            self.d["effectiveTime"] = dateutil.parser.parse(
                e["effectiveTime"]
            ).timestamp()
            if e["payload"]["__typename"] == "BakerStakeThresholdChainUpdatePayload":
                self.d["payload"] = {}
                self.d["payload"]["update"] = {
                    "minimumThresholdForBaking": e["payload"]["amount"]
                }
                self.d["payload"]["updateType"] = "bakerStakeThreshold"

            elif e["payload"]["__typename"] == "ProtocolChainUpdatePayload":
                self.d["payload"] = {}
                self.d["payload"]["update"] = {
                    "specificationUrl": e["payload"]["specificationUrl"],
                    "message": e["payload"]["message"],
                    "specificationHash": e["payload"]["specificationHash"],
                    "specificationAuxiliaryData": e["payload"][
                        "specificationAuxiliaryDataAsHex"
                    ],
                }
                self.d["payload"]["updateType"] = "protocol"
            else:
                self.d["payload"] = e["payload"]

        elif etype == TransactionType.ContractInitialized.name:
            self.d["address"] = {
                "index": e["contractAddress"]["index"],
                "subindex": e["contractAddress"]["subIndex"],
            }
            self.d["amount"] = e["amount"]
            self.d["contractVersion"] = 0  # note: not available in QL
            self.d["initName"] = e["initName"]
            self.d["ref"] = e["moduleRef"]
            self.d["events"] = e["HEX"]["nodes"]

        elif etype == TransactionType.ContractResumed.name:
            self.d["address"] = {
                "index": e["contractAddress"]["index"],
                "subindex": e["contractAddress"]["subIndex"],
            }
            self.d["success"] = e["success"]

        elif etype == TransactionType.ContractInterrupted.name:
            self.d["address"] = {
                "index": e["contractAddress"]["index"],
                "subindex": e["contractAddress"]["subIndex"],
            }
            self.d["events"] = e["hex_events"]

        elif etype == TransactionType.ContractUpdated.name:
            self.d["address"] = {
                "index": e["contractAddress"]["index"],
                "subindex": e["contractAddress"]["subIndex"],
            }
            self.d["amount"] = e["amount"]
            self.d["contractVersion"] = 0  # note: not available in QL
            self.d["instigator"] = self.determineAddress(e["instigator"])
            self.d["message"] = e["messageAsHex"]
            self.d["receiveName"] = e["receiveName"]
            self.d["events"] = e["eventsAsHex"]["nodes"]

        elif etype == TransactionType.ContractModuleDeployed.name:
            self.d["contents"] = e["moduleRef"]

        elif etype == TransactionType.CredentialDeployed.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["regId"] = e["regId"]

        elif etype == TransactionType.CredentialsUpdated.name:
            self.d["account"] = e["accountAddress"]["asString"]
            self.d["newCredIds"] = e["newCredIds"]
            self.d["newThreshold"] = e["newThreshold"]
            self.d["removedCredIds"] = e["removedCredIds"]

        elif etype == TransactionType.DataRegistered.name:
            self.d["data"] = e["dataAsHex"]

        elif etype == TransactionType.EncryptedAmountsRemoved.name:
            pass

        elif etype == TransactionType.NewEncryptedAmount.name:
            pass

        elif etype == TransactionType.TransferMemo.name:
            # note this is actually NOT the same as the node, as the node does NOT decode.
            self.d["memo"] = e["decoded"]["text"]

        elif etype == TransactionType.Transferred.name:
            self.d["amount"] = int(e["amount"])
            self.d["from"] = self.determineAddress(e["from"])
            self.d["to"] = self.determineAddress(e["to"])

        elif etype == TransactionType.TransferredWithSchedule.name:
            self.d["from"] = e["fromAccountAddress"]["asString"]
            self.d["to"] = e["toAccountAddress"]["asString"]
            self.d["amount"] = [
                (dateutil.parser.parse(x["timestamp"]).timestamp() * 1000, x["amount"])
                for x in e["amountsSchedule"]["nodes"]
            ]

        return self.d


class Transaction:
    """
    Canonical Transaction. Transactions from either the node or GraphQL will be morphed to fit this class
    Node language will be the go-to point, so graphQL terms are translated to fit the node.
    """

    def __init__(self, node):
        self.node = node
        self.block: ConcordiumBlockInfo

    def find_memo_and_amount(self):
        self.memo = None
        self.amount = None
        if (self.contents == "transferWithMemo") or (self.contents == "Transferred"):
            if "events" in self.result:
                for event in self.result.get("events", []):
                    if event["tag"] == TransactionType.TransferMemo.name:
                        self.memo = event["memo"]
                    if event["tag"] == TransactionType.Transferred.name:
                        self.amount = event["amount"]

        return self

    def get_domain_from_cache_or_node(self):
        if self.cns_domain.tokenId in self.node.cns_cache_by_token_id:
            self.cns_domain.domain_name = self.node.cns_cache_by_token_id[
                self.cns_domain.tokenId
            ]
            console.log(f"Using cache for {self.cns_domain.domain_name}")
        else:
            self.cns_domain.get_cns_domain_name(self.node, self.cns_domain.tokenId)

            self.node.cns_cache_by_token_id[self.cns_domain.tokenId] = (
                self.cns_domain.domain_name
            )
            self.node.cns_cache_by_name[self.cns_domain.domain_name] = (
                self.cns_domain.tokenId
            )
            self.node.save_cns_cache()
            console.log(f"Saving cache for {self.cns_domain.domain_name}")

    def set_possible_cns_domain(self):
        self.cns_domain = CNSDomain()
        if "events" in self.result:
            for event in self.result.get("events", []):
                if event["tag"] in [
                    TransactionType.ContractInterrupted.name,
                    "Interrupted",
                ]:
                    # console.log(f'{event["tag"]=}')
                    if "events" in event:
                        if len(event["events"]) > 0:
                            the_event = event["events"][0]

                            (
                                tag_,
                                contract_index,
                                contract_subindex,
                                token_id_,
                                seller_,
                                winner_,
                                price_,
                                seller_share,
                                royalty_length_,
                                royalties,
                                owner_,
                                bidder_,
                                amount_,
                            ) = self.cns_domain.finalize(the_event)

                            # tag_, contract_index, contract_subindex, \
                            # token_id_, bidder_, amount_ = self.cns_domain.bidEvent(the_event)

                            self.cns_domain.tokenId = token_id_
                            # self.cns_domain.amount = amount_

                            self.get_domain_from_cache_or_node()
                # console.log(f"{event['tag']=}")
                if event["tag"] in [TransactionType.ContractUpdated.name, "Updated"]:
                    # console.log(f'{event["receiveName"]=}')
                    self.cns_domain.function_calls[event["receiveName"]] = event[
                        "message"
                    ]

                    if event["receiveName"] == "BictoryCnsNft.transfer":
                        self.cns_domain.action = CNSActions.transfer
                        (
                            self.cns_domain.tokenId,
                            self.cns_domain.transfer_to,
                        ) = self.cns_domain.decode_transfer_to_from(event["message"])

                        self.get_domain_from_cache_or_node()

                    if event["receiveName"] == "BictoryCns.register":
                        self.cns_domain.amount = event["amount"]
                        self.cns_domain.action = CNSActions.register
                        (
                            self.cns_domain.domain_name,
                            self.cns_domain.register_address,
                            self.cns_domain.duration_years,
                        ) = self.cns_domain.decode_from_register(event["message"])

                    if event["receiveName"] == "BictoryCns.extend":
                        self.cns_domain.amount = event["amount"]
                        self.cns_domain.action = CNSActions.extend
                        (
                            self.cns_domain.domain_name,
                            self.cns_domain.duration_years,
                        ) = self.cns_domain.decode_from_extend(event["message"])

                    if event["receiveName"] == "BictoryCns.createSubdomain":
                        self.cns_domain.action = CNSActions.createSubdomain
                        self.cns_domain.amount = event["amount"]
                        self.cns_domain.subdomain = (
                            self.cns_domain.decode_subdomain_from(event["message"])
                        )

                    if event["receiveName"] == "BictoryCns.setAddress":
                        self.cns_domain.action = CNSActions.setAddress
                        (
                            self.cns_domain.domain_name,
                            self.cns_domain.set_address,
                        ) = self.cns_domain.decode_set_address_from(event["message"])

                    if event["receiveName"] == "BictoryCns.setData":
                        self.cns_domain.action = CNSActions.setData
                        (
                            self.cns_domain.domain_name,
                            self.cns_domain.set_data_key,
                            self.cns_domain.set_data_value,
                        ) = self.cns_domain.decode_set_data_from(event["message"])

                    if event["receiveName"] == "BictoryNftAuction.bid":
                        self.cns_domain.action = CNSActions.bid

                        if len(event["events"]) > 0:
                            the_event = event["events"][0]

                            (
                                tag_,
                                contract_index,
                                contract_subindex,
                                token_id_,
                                bidder_,
                                amount_,
                            ) = self.cns_domain.bidEvent(the_event)

                            self.cns_domain.tokenId = token_id_
                            self.cns_domain.amount = amount_

                            self.get_domain_from_cache_or_node()

                    if event["receiveName"] == "BictoryNftAuction.finalize":
                        self.cns_domain.action = CNSActions.finalize

                        if len(event["events"]) > 0:
                            the_event = event["events"][0]

                            (
                                tag_,
                                contract_index,
                                contract_subindex,
                                token_id_,
                                seller_,
                                winner_,
                                price_,
                                seller_share,
                                royalty_length_,
                                royalties,
                                owner_,
                                bidder_,
                                amount_,
                            ) = self.cns_domain.finalize(the_event)

                            self.cns_domain.tokenId = token_id_

                            self.get_domain_from_cache_or_node()

                    if event["receiveName"] == "BictoryNftAuction.cancel":
                        self.cns_domain.action = CNSActions.cancel

                        if len(event["events"]) > 0:
                            the_event = event["events"][0]

                            (
                                tag_,
                                contract_index,
                                contract_subindex,
                                token_id_,
                                owner_,
                            ) = self.cns_domain.cancelEvent(the_event)

                            self.cns_domain.tokenId = token_id_

                            self.get_domain_from_cache_or_node()

                    if event["receiveName"] == "BictoryCnsNft.getTokenExpiry":
                        self.cns_domain.action = CNSActions.getTokenExpiry
                        self.cns_domain.tokenId = self.cns_domain.decode_token_id_from(
                            event["message"]
                        )

                        self.get_domain_from_cache_or_node()

    def decode_possible_memo(self):
        _events = []
        if "events" in self.result:
            for event in self.result.get("events", []):
                Event(event).translate_memo_if_present()

    def init_from_node(self, t):
        # note that only blockHash, blockHeight, blockSlotTime, finalized is used
        self.block: ConcordiumBlockInfo = t.get("blockInfo", None)
        if not self.block:
            self.block = {}
            self.block["blockHash"] = t["hash"]
            self.block["blockHeight"] = t["blockHeight"]
            self.block["blockSlotTime"] = t["blockSlotTime"]
        self.cost = t["cost"]
        self.energyCost = t["energyCost"]
        self.hash = t["hash"]
        self.index = t["index"]
        if t["result"]["outcome"] == "success":
            self.result = {
                "events": t["result"]["events"],
                "outcome": t["result"]["outcome"],
            }
        else:
            self.result = {"outcome": t["result"]["outcome"]}
        self.sender = t["sender"]
        self.type = t["type"]["type"]  # accountTransaction
        self.contents = t["type"]["contents"]  # transfer
        self.decode_possible_memo()
        self.set_possible_cns_domain()
        # self.node = None
        # console.log(f"{self.hash=} | {self.cns_domain.domain_name=}")
        return self

    def init_from_mongo_tx(self, t):
        # note that only blockHash, blockHeight, blockSlotTime, finalized is used
        self.block = {}
        self.block["blockHash"] = t["blockHash"]
        self.block["blockHeight"] = t["blockHeight"]
        self.block["blockSlotTime"] = t["blockSlotTime"]
        self.cost = t["cost"]
        self.energyCost = t["energyCost"]
        self.hash = t["hash"]
        self.index = t["index"]
        if t["result"]["outcome"] == "success":
            self.result = {
                "events": t["result"]["events"],
                "outcome": t["result"]["outcome"],
            }
        else:
            self.result = {"outcome": t["result"]["outcome"]}
        self.sender = t["sender"]
        self.type = t["type"]["type"]  # accountTransaction
        self.contents = t["type"]["contents"]  # transfer
        self.decode_possible_memo()
        self.set_possible_cns_domain()
        # self.node = None
        # console.log(f"{self.hash=} | {self.cns_domain.domain_name=}")
        return self

    # Bot classification requirements
    def get_sum_amount_from_scheduled_transfer(self, events):
        sum = 0
        for event in events[0]["amount"]:
            sum += int(event[1])
        return sum

    def classify_transaction_for_bot(self):
        # console.log(f"BOT: {self.hash=} | {self.cns_domain.domain_name=}")
        t = self
        result = ClassificationResult()
        result.domain_name = self.cns_domain.domain_name
        result.accounts_involved_all = True
        result.type = t.type
        result.contents = t.contents
        result.sender = t.sender
        result.txHash = t.hash
        result.receiver = None

        if self.type == "credentialDeploymentTransaction":
            if t.result["outcome"] == "success":
                events = t.result["events"]

                for event in events:
                    if event["tag"] == "AccountCreated":
                        result.sender = event["contents"]

        if self.type == "accountTransaction":
            if t.result["outcome"] == "success":
                events = t.result["events"]

                for event in events:
                    if event["tag"] == "Transferred":
                        result.receiver = event["to"]["address"]
                        result.amount = event["amount"]
                        result.accounts_involved_transfer = True

                    elif event["tag"] == "TransferredWithSchedule":
                        result.receiver = event["to"]
                        result.amount = self.get_sum_amount_from_scheduled_transfer(
                            events
                        )
                        result.accounts_involved_transfer = True

                    elif event["tag"] == "ContractInitialized":
                        result.contract = {
                            "tag": event["tag"],
                            "index": event["address"]["index"],
                            "subindex": event["address"]["subindex"],
                        }

                        result.list_of_contracts_involved.add(
                            (event["address"]["index"], event["address"]["subindex"])
                        )
                        result.contracts_involved = True

                    elif event["tag"] == "Updated":
                        result.contract = {
                            "tag": event["tag"],
                            "index": event["address"]["index"],
                            "subindex": event["address"]["subindex"],
                        }

                        result.list_of_contracts_involved.add(
                            (event["address"]["index"], event["address"]["subindex"])
                        )
                        result.contracts_involved = True

                    elif event["tag"] == "Resumed":
                        result.contract = {
                            "tag": event["tag"],
                            "index": event["address"]["index"],
                            "subindex": event["address"]["subindex"],
                        }

                        result.list_of_contracts_involved.add(
                            (event["address"]["index"], event["address"]["subindex"])
                        )
                        result.contracts_involved = True

                    elif event["tag"] == "Interrupted":
                        result.contract = {
                            "tag": event["tag"],
                            "index": event["address"]["index"],
                            "subindex": event["address"]["subindex"],
                        }

                        result.list_of_contracts_involved.add(
                            (event["address"]["index"], event["address"]["subindex"])
                        )
                        result.contracts_involved = True

        if result.amount:
            result.amount = int(result.amount) / 1_000_000
        return result, self

    def request_delegated_amount_for_account_on_previous_block(
        self, account_id, block_height
    ):
        previous_block_hash = self.node.request_block_hash_at_height(block_height - 1)
        account_info_at_previous_block = self.node.request_accountInfo_at(
            previous_block_hash, account_id
        )
        if "accountDelegation" in account_info_at_previous_block:
            previous_amount = account_info_at_previous_block["accountDelegation"][
                "stakedAmount"
            ]
        else:
            previous_amount = None
        return previous_amount

    def request_staked_amount_for_account_on_previous_block(
        self, account_id, block_height
    ):
        previous_block_hash = self.node.request_block_hash_at_height(block_height - 1)
        account_info_at_previous_block = self.node.request_accountInfo_at(
            previous_block_hash, account_id
        )
        if "accountBaker" in account_info_at_previous_block:
            previous_amount = account_info_at_previous_block["accountBaker"][
                "stakedAmount"
            ]
        else:
            previous_amount = None
        return previous_amount

    def classify_transaction_for_bot_for_delegation_and_staking(self):
        t = self
        result = DelegationAndStakingResult()

        if self.type == "accountTransaction":
            result.sender = t.sender
            result.txHash = t.hash

            if t.result["outcome"] == "success":
                events = t.result["events"]

                for event in events:
                    perc = ""
                    # print (f"event = {event}")
                    if event["tag"] in ["BakerStakeDecreased"]:
                        result.bakerId = event["bakerId"]
                        previous_amount = (
                            self.request_staked_amount_for_account_on_previous_block(
                                result.sender, self.block.blockHeight
                            )
                        )
                        if previous_amount:
                            if int(previous_amount) > 0:
                                result.perc = f" ({(100*(int(previous_amount) - int(event['newStake']))/int(previous_amount)):,.2f}%)"
                                result.unstaked_amount = (
                                    int(previous_amount) - int(event["newStake"])
                                ) / 1_000_000

                    if event["tag"] in [
                        "DelegationStakeIncreased",
                        "DelegationStakeDecreased",
                    ]:
                        previous_amount = (
                            self.request_delegated_amount_for_account_on_previous_block(
                                result.sender, self.block.blockHeight
                            )
                        )
                        if previous_amount:
                            if int(previous_amount) > 0:
                                if int(event["newStake"]) > int(previous_amount):
                                    result.perc = f" ({(100*((int(event['newStake'])/int(previous_amount)) - 1)):,.2f}%)"
                                else:
                                    result.perc = f" ({(100*(int(previous_amount) - int(event['newStake']))/int(previous_amount)):,.2f}%)"

                    if event["tag"] == "DelegationSetDelegationTarget":
                        if event["delegationTarget"]["delegateType"] == "Baker":
                            result.bakerId = event["delegationTarget"]["bakerId"]
                            result.message += (
                                f"Delegation target set to {result.bakerId:,.0f}. "
                            )

                    elif event["tag"] == "DelegationStakeIncreased":
                        if previous_amount:
                            result.message += f"Stake increased{result.perc} to {(int(event['newStake'])/1_000_000):,.0f} CCD."
                        else:
                            result.message += f"Stake set to {(int(event['newStake'])/1_000_000):,.0f} CCD."
                        result.request_target = True

                    elif event["tag"] == "DelegationStakeDecreased":
                        result.message += f"Stake decreased{result.perc} to {(int(event['newStake'])/1_000_000):,.0f} CCD."
                        result.request_target = True

                    elif event["tag"] == "BakerStakeDecreased":
                        result.message += f"Baker stake decreased {result.unstaked_amount:,.0f} CCD{result.perc} to {(int(event['newStake'])/1_000_000):,.0f} CCD."
                        result.request_target = False
                        result.unstaked = True

                    elif event["tag"] == "DelegationRemoved":
                        result.message += f"Delegation removed."
                        result.request_target = True

        return result

    def define_type_and_contents_from_graphQL(self, t):
        _type = t["transactionType"]["__typename"]  # AccountTransaction

        if _type == "AccountTransaction":
            try:
                _contents = t["transactionType"]["accountTransactionType"]
            except:
                _contents = "UNKNOWN"
        elif _type == "CredentialDeploymentTransaction":
            try:
                _contents = t["transactionType"]["credentialDeploymentTransactionType"]
            except:
                _contents = "UNKNOWN"
        elif _type == "UpdateTransaction":
            try:
                _contents = t["transactionType"]["updateTransactionType"]
            except:
                _contents = "UNKNOWN"
        else:
            _contents = "UNKNOWN"

        # now translate to node language
        # console.log(f"{_type=}, {_contents=}")
        self.type = TransactionTypeFromQLToNode[_type].value
        if t["result"]["__typename"] == "Success":
            self.contents = TransactionContentsFromQLToNode[_contents].value
        else:
            self.contents = None

    def translate_result_from_graphQL(self, t):
        if t["result"]["__typename"] == "Success":
            _outcome = "success"
        elif t["result"]["__typename"] == "Rejected":
            _outcome = "reject"
            _rejectReason = t["result"]["reason"]["__typename"]

        _events = []
        if "events" in t["result"]:
            for event in t["result"]["events"]["nodes"]:
                e = Event(event).translate_event_from_graphQL()
                _events.append(e)

        if _outcome == "success":
            return {"events": _events, "outcome": _outcome}
        else:
            return {
                "events": _events,
                "outcome": _outcome,
                "rejectReason": _rejectReason,
            }

    def init_from_graphQL(self, t):
        # try:
        #     t['block']['blockSlotTime'] = dateutil.parser.parse(t['block']['blockSlotTime'])
        # except:
        #     pass
        self.block = t["block"]
        self.cost = t["ccdCost"]
        self.energyCost = t["energyCost"]
        self.hash = t["transactionHash"]
        self.index = t["transactionIndex"]
        self.sender = (
            t["senderAccountAddress"]["asString"]
            if t["senderAccountAddress"] is not None
            else None
        )
        self.define_type_and_contents_from_graphQL(t)
        self.result = self.translate_result_from_graphQL(t)
        self.set_possible_cns_domain()
        self.node = None
        return self



# ruff: noqa: F403, F405, E402
from __future__ import annotations
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import (
    CCD_RawQuorumCertificate,
    CCD_RoundStatus,
)
from ccdexplorer_fundamentals.GRPCClient.types_pb2 import *
from ccdexplorer_fundamentals.enums import NET
from enum import Enum
from ccdexplorer_fundamentals.GRPCClient.queries._SharedConverters import (
    Mixin as _SharedConverters,
)
import os
import sys

sys.path.append(os.path.dirname("ccdexplorer_fundamentals"))
from ccdexplorer_fundamentals.GRPCClient.CCD_Types import *
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from ccdexplorer_fundamentals.GRPCClient import GRPCClient


class Mixin(_SharedConverters):
    def convertRoundTimeout(self, message) -> CCD_RoundTimeout:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) is RawTimeoutCertificate:
                result[key] = self.convertRawTimeoutCertificate(value)
            elif type(value) is RawQuorumCertificate:
                result[key] = self.convertRawQuorumCertificate(value)

        return CCD_RoundTimeout(**result)

    def convertSignatories(self, message) -> list[CCD_FinalizerIndex]:
        signatories = []
        for finalizer_index in message:
            signatories.append(finalizer_index.value)

        return signatories

    def convertFinalizers(self, message) -> list[CCD_BakerId]:
        finalizers = []
        for baker_id in message:
            finalizers.append(baker_id.value)

        return finalizers

    def convertFullBakerInfo(self, message) -> list[CCD_FullBakerInfo]:
        bakers = []
        for baker in message:
            result = {}
            for field, value in baker.ListFields():
                key = field.name
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)

            bakers.append(CCD_FullBakerInfo(**result))
        return bakers

    def convertRawFinalizerRounds(self, message) -> list[CCD_RawFinalizerRound]:
        rounds = []
        for round in message:
            result = {}
            for field, value in round.ListFields():
                key = field.name
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)
                elif key == "finalizers":
                    result[key] = self.convertSignatories(value)
            rounds.append(CCD_RawFinalizerRound(**result))
        return rounds

    def convertRawTimeoutCertificate(self, message) -> CCD_RawTimeoutCertificate:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif key in ["qc_rounds_first_epoch", "qc_rounds_second_epoch"]:
                result[key] = self.convertRawFinalizerRounds(value)

        return CCD_RawTimeoutCertificate(**result)

    def convertRawQuorumCertificate(self, message) -> CCD_RawQuorumCertificate:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif key == "signatories":
                result[key] = self.convertSignatories(value)

        return CCD_RawQuorumCertificate(**result)

    def convertRoundStatus(self, message) -> CCD_RoundStatus:
        result = {}
        for field, value in message.ListFields():
            key = field.name
            if self.valueIsEmpty(value):
                pass
            else:
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)
                elif type(value) is RawQuorumCertificate:
                    result[key] = self.convertRawQuorumCertificate(value)
                elif type(value) is RoundTimeout:
                    result[key] = self.convertRoundTimeout(value)

        return CCD_RoundStatus(**result)

    def convertQuorumMessage(self, message) -> CCD_QuorumMessage:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_QuorumMessage(**result)

    def convertTimeoutMessage(self, message) -> CCD_TimeoutMessage:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif type(value) is RawQuorumCertificate:
                result[key] = self.convertRawQuorumCertificate(value)

        return CCD_TimeoutMessage(**result)

    def convertPersistentRoundStatus(self, message) -> CCD_PersistentRoundStatus:
        result = {}
        for field, value in message.ListFields():
            key = field.name
            if self.valueIsEmpty(value):
                pass
            else:
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)
                elif type(value) is QuorumMessage:
                    result[key] = self.convertQuorumMessage(value)
                elif type(value) is TimeoutMessage:
                    result[key] = self.convertTimeoutMessage(value)
                elif type(value) is RawQuorumCertificate:
                    result[key] = self.convertRawQuorumCertificate(value)

        return CCD_PersistentRoundStatus(**result)

    def convertLiveBlocks(self, message) -> list[CCD_BlockHash]:
        live_blocks = []
        for hash in message:
            live_blocks.append(self.convertType(hash))

        return live_blocks

    def convertBlocksAtBranchHeight(self, message) -> list[CCD_BlockHash]:
        blocks_at_branch_height = []
        for hash in message:
            blocks_at_branch_height.append(self.convertType(hash))

        return blocks_at_branch_height

    def convertBranches(self, message) -> list[CCD_BranchBlocks]:
        branches = []
        for hash in message:
            for field, value in hash.ListFields():
                key = field.name
                if key == "blocks_at_branch_height":
                    blocks_at_branch_height = self.convertBlocksAtBranchHeight(value)
                branches.append(
                    CCD_BranchBlocks(
                        **{"blocks_at_branch_height": blocks_at_branch_height}
                    )
                )

        return branches

    def convertRoundExistingBlock(self, message) -> CCD_RoundExistingBlock:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_RoundExistingBlock(**result)

    def convertRoundExistingBlocks(self, message) -> list[CCD_RoundExistingBlock]:
        round_existing_blocks = []
        for value in message:
            round_existing_blocks.append(self.convertRoundExistingBlock(value))

        return round_existing_blocks

    def convertRoundExistingQC(self, message) -> CCD_RoundExistingQC:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)

        return CCD_RoundExistingQC(**result)

    def convertRoundExistingQCs(self, message) -> list[CCD_RoundExistingQC]:
        round_existing_qcs = []
        for value in message:
            round_existing_qcs.append(self.convertRoundExistingQC(value))

        return round_existing_qcs

    def convertBlockTableSummary(self, message) -> CCD_BlockTableSummary:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            if key == "live_blocks":
                result[key] = self.convertLiveBlocks(value)

        return CCD_BlockTableSummary(**result)

    def convertRawFinalizationEntry(self, message) -> CCD_RawFinalizationEntry:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif type(value) is RawQuorumCertificate:
                result[key] = self.convertRawQuorumCertificate(value)

        return CCD_RawFinalizationEntry(**result)

    def convertBakersAndFinalizers(self, message) -> CCD_BakersAndFinalizers:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif key == "finalizers":
                result[key] = self.convertFinalizers(value)
            elif key == "bakers":
                result[key] = self.convertFullBakerInfo(value)

        return CCD_BakersAndFinalizers(**result)

    def convertEpochBakers(self, message) -> CCD_EpochBakers:
        result = {}
        for field, value in message.ListFields():
            key = field.name
            if self.valueIsEmpty(value):
                pass
            else:
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)
                elif type(value) is BakersAndFinalizers:
                    result[key] = self.convertBakersAndFinalizers(value)

        return CCD_EpochBakers(**result)

    def convertTimeoutMessages(self, message) -> CCD_TimeoutMessages:
        result = {}
        for field, value in message.ListFields():
            key = field.name

            if type(value) in self.simple_types:
                result[key] = self.convertType(value)
            elif key in ["first_epoch_timeouts", "second_epoch_timeouts"]:
                result[key] = self.convertTimeoutMessage(value)

        return CCD_TimeoutMessages(**result)

    def get_consensus_detailed_status(
        self: GRPCClient,
        gen_index: Optional[int] = None,
        net: Enum = NET.MAINNET,
    ) -> CCD_ConsensusDetailedStatus:
        result = {}
        consensus_detailed_status_query = (
            self.generate_consensus_detailed_status_query()
        )

        grpc_return_value: ConsensusDetailedStatus = self.stub_on_net(
            net, "GetConsensusDetailedStatus", consensus_detailed_status_query
        )

        for field, value in grpc_return_value.ListFields():
            key = field.name
            if self.valueIsEmpty(value):
                pass
            else:
                if type(value) in self.simple_types:
                    result[key] = self.convertType(value)
                elif type(value) is PersistentRoundStatus:
                    result[key] = self.convertPersistentRoundStatus(value)
                elif type(value) is RoundStatus:
                    result[key] = self.convertRoundStatus(value)
                elif type(value) is BlockTableSummary:
                    result[key] = self.convertBlockTableSummary(value)
                elif key == "branches":
                    result[key] = self.convertBranches(value)
                elif key == "round_existing_blocks":
                    result[key] = self.convertRoundExistingBlocks(value)
                elif key == "round_existing_qcs":
                    result[key] = self.convertRoundExistingQCs(value)
                elif type(value) is RawFinalizationEntry:
                    result[key] = self.convertRawFinalizationEntry(value)
                elif type(value) is EpochBakers:
                    result[key] = self.convertEpochBakers(value)
                elif type(value) is TimeoutMessages:
                    result[key] = self.convertTimeoutMessages(value)

        return CCD_ConsensusDetailedStatus(**result)
